import logging
import numpy as np
import scipy.sparse as sprs
import scipy.spatial as sptl
from openpnm.core import Domain
from openpnm import topotools
from openpnm.utils import Docorator
from openpnm.utils import Workspace
import openpnm.models.network as mods
logger = logging.getLogger(__name__)
ws = Workspace()
docstr = Docorator()


@docstr.get_sections(base='NetworkSettings', sections=['Parameters'])
@docstr.dedent
class NetworkSettings:
    r"""

    Parameters
    ----------
    %(BaseSettings.parameters)s
    """


@docstr.get_sections(base='Network', sections=['Parameters'])
@docstr.dedent
class Network(Domain):
    r"""
    This class contains the main functionality used by all networks.

    Parameters
    ----------
    settings : dataclass-like or dict, optional
        User defined settings for the object to override defaults. Can be a
        dataclass-type object with settings stored as attributes or a python
        dicionary of key-value pairs. Settings are stored in the ``settings``
        attribute of the object.
    name : str, optional
        A unique name to assign to the object for easier identification. If
        not given one will be generated.

    coords : array_like (optional)
        An Np-by-3 array of [x, y, z] coordinates for each pore
    conns : array_like (optional)
        An Nt-by-2 array of [head, tail] connections between pores

    Examples
    --------
    >>> import openpnm as op
    >>> import numpy as np

    Create some pore coordinates and connections manually and assign to a
    Network instance. Consider a linear network of 4 pores and 3
    throats:

    ::

        0 ―― 1 ―― 3 ―― 2

    >>> coords = np.array([[0, 0, 0], [1, 0, 0], [2, 0, 0], [3, 0, 0]])
    >>> conns = np.array([[0, 1], [1, 3], [2, 3]])
    >>> pn = op.network.Network(conns=conns, coords=coords)

    Networks have two required properties: 'pore.coords' and
    'throat.conns'. These arrays indicate the spatial location of each
    pore, and which pores are connected to which. Without these the
    network object cannot function.

    >>> print(pn.props())
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1     : pore.coords
    2     : throat.conns
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    The ``Network`` class has several methods for querying the topology.

    >>> Ps = pn.find_neighbor_pores(pores=1)
    >>> print(Ps)
    [0 3]
    >>> Ts = pn.find_neighbor_throats(pores=[0, 1])
    >>> print(Ts)
    [0 1]
    >>> print(pn.num_neighbors(2))
    [1]

    All of the topological queries are accomplished by inspecting the
    adjacency and incidence matrices. They are created on demand, and are
    stored for future use to save construction time.

    """

    def __init__(self, conns=None, coords=None, name='net', **kwargs):
        super().__init__(name=name, **kwargs)
        self.settings._update(NetworkSettings())
        self._am = {}
        self._im = {}

        if coords is not None:
            coords = np.array(coords)
            self['pore.coords'] = coords

        if conns is not None:
            conns = np.array(conns)
            self['throat.conns'] = conns

        self.add_model(propname='pore.coordination_number',
                       model=mods.coordination_number,
                       regen_mode='deferred',
                       domain='all')
        self.add_model(propname='throat.spacing',
                       model=mods.pore_to_pore_distance,
                       regen_mode='deferred',
                       domain='all')

    def _post_init(self, *args, **kwargs):
        # A collection of functions to run after initialization of a network
        topotools.label_faces(self, label=None)
        drop = ['pore.left', 'pore.right', 'pore.front', 'pore.back',
                'pore.bottom', 'pore.top']
        [self.pop(item) for item in list(self.keys()) if item in drop]

    def __setitem__(self, key, value):
        if key == 'throat.conns':
            if np.any(value[:, 0] > value[:, 1]):
                logger.warning('Converting throat.conns to be upper triangular')
                value = np.sort(value, axis=1)
        super().__setitem__(key, value)

    def get_adjacency_matrix(self, fmt='coo'):
        r"""
        Adjacency matrix in the specified sparse format, with throat IDs
        indicating the non-zero values.

        Parameters
        ----------
        fmt : str, default is 'coo'
            The sparse storage format to return. Options are:
                **'coo'** : This is the native format of OpenPNM's data
                **'lil'** : Enables row-wise slice of the matrix
                **'csr'** : Favored by most linear algebra routines
                **'dok'** : Enables subscript access of locations

        Notes
        -----
        This method will only create the requested matrix in the specified
        format if one is not already saved on the object.  If not present,
        this method will create and return the matrix, as well as store it
        for future use.

        To obtain a matrix with weights other than throat IDs at each
        non-zero location use ``create_adjacency_matrix``.

        To obtain the non-directed graph, with only upper-triangular
        entries, use ``sp.sparse.triu(am, k=1)``.

        """
        # Retrieve existing matrix if available
        if fmt in self._am.keys():
            am = self._am[fmt]
        else:
            am = self.create_adjacency_matrix(weights=self.Ts, fmt=fmt)
            self._am[fmt] = am
        return am

    def get_incidence_matrix(self, fmt='coo'):
        r"""
        Incidence matrix in the specified sparse format, with pore IDs
        indicating the non-zero values.

        Parameters
        ----------
        fmt : str, default is 'coo'
            The sparse storage format to return. Options are:
                **'coo'** : This is the native format of OpenPNM's data
                **'lil'** : Enables row-wise slice of the matrix
                **'csr'** : Favored by most linear algebra routines
                **'dok'** : Enables subscript access of locations

        Notes
        -----
        This method will only create the requested matrix in the specified
        format if one is not already saved on the object. If not present,
        this method will create and return the matrix, as well as store it
        for future use.

        To obtain a matrix with weights other than pore IDs at each
        non-zero location use ``create_incidence_matrix``.

        """
        if fmt in self._im.keys():
            im = self._im[fmt]
        elif self._im.keys():
            im = self._im[list(self._im.keys())[0]]
            tofmt = getattr(im, f"to{fmt}")
            im = tofmt()
            self._im[fmt] = im
        else:
            im = self.create_incidence_matrix(weights=self.Ts, fmt=fmt)
            self._im[fmt] = im
        return im

    im = property(fget=get_incidence_matrix)

    am = property(fget=get_adjacency_matrix)

    def create_adjacency_matrix(self, weights=None, fmt='coo', triu=False,
                                drop_zeros=False):
        r"""
        Generates a weighted adjacency matrix in the desired sparse format

        Parameters
        ----------
        weights : array_like, optional
            An array containing the throat values to enter into the matrix
            (in graph theory these are known as the 'weights').

            If the array is Nt-long, it implies that the matrix is
            symmetric, so the upper and lower triangular regions are
            mirror images. If it is 2*Nt-long then it is assumed that the
            first Nt elements are for the upper triangle, and the last Nt
            element are for the lower triangular.

            If omitted, ones are used to create a standard adjacency
            matrix representing connectivity only.
        fmt : str, optional
            The sparse storage format to return. Options are:
                **'coo'** : This is the native format of OpenPNM's data
                **'lil'** : Enables row-wise slice of the matrix
                **'csr'** : Favored by most linear algebra routines
                **'dok'** : Enables subscript access of locations
        triu : bool, default is ``False``
            If ``True``, the returned sparse matrix only contains the
            upper-triangular elements. This argument is ignored if the
            ``weights`` array is 2*Nt-long.
        drop_zeros : bool, default is ``False``
            If ``True``, applies the ``eliminate_zeros`` method of the
            sparse array to remove all zero locations.

        Returns
        -------
        An adjacency matrix in the specified Scipy sparse format.

        Notes
        -----
        The adjacency matrix is used by OpenPNM for finding the pores
        connected to a give pore or set of pores. Specifically, an
        adjacency matrix has Np rows and Np columns. Each row represents
        a pore, containing non-zero values at the locations corresponding
        to the indices of the pores connected to that pore. The
        ``weights`` argument indicates what value to place at each
        location, with the default being 1's to simply indicate
        connections. Another useful option is throat indices, such that
        the data values on each row indicate which throats are connected
        to the pore.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> weights = np.random.rand(pn.num_throats(), ) < 0.5
        >>> am = pn.create_adjacency_matrix(weights=weights, fmt='csr')

        """
        allowed_weights = [(self.Nt,), (2 * self.Nt,), (self.Nt, 2)]
        # Check if provided data is valid
        if weights is None:
            weights = np.ones((self.Nt,), dtype=int)
        elif np.shape(weights) not in allowed_weights:
            raise Exception('Received weights are of incorrect length')
        weights = np.array(weights)

        # Append row & col to each other, and data to itself
        conn = self['throat.conns']
        row = conn[:, 0]
        col = conn[:, 1]
        if weights.shape == (2 * self.Nt):
            # The flip is necessary since we want [conns.T, reverse(conns).T].T
            row = np.append(row, conn[:, 1])
            col = np.append(col, conn[:, 0])
        elif weights.shape == (self.Nt, 2):
            # The flip is necessary since we want [conns.T, reverse(conns).T].T
            row = np.append(row, conn[:, 1])
            col = np.append(col, conn[:, 0])
            weights = weights.flatten(order='F')
        elif not triu:
            # The flip is necessary since we want [conns.T, reverse(conns).T].T
            row = np.append(row, conn[:, 1])
            col = np.append(col, conn[:, 0])
            weights = np.append(weights, weights)

        # Generate sparse adjacency matrix in 'coo' format
        temp = sprs.coo_matrix((weights, (row, col)), (self.Np, self.Np))

        if drop_zeros:
            temp.eliminate_zeros()

        # Convert to requested format
        if fmt == 'coo':
            pass  # temp is already in coo format
        elif fmt == 'csr':
            temp = temp.tocsr()
        elif fmt == 'lil':
            temp = temp.tolil()
        elif fmt == 'dok':
            temp = temp.todok()

        return temp

    def create_incidence_matrix(self, weights=None, fmt='coo',
                                drop_zeros=False):
        r"""
        Creates a weighted incidence matrix in the desired sparse format

        Parameters
        ----------
        weights : array_like, optional
            An array containing the throat values to enter into the matrix
            (in graph theory these are known as the 'weights'). If
            omitted, ones are used to create a standard incidence matrix
            representing connectivity only.
        fmt : str, default is 'coo'
            The sparse storage format to return. Options are:
                **'coo'** : This is the native format of OpenPNM's data
                **'lil'** : Enables row-wise slice of the matrix
                **'csr'** : Favored by most linear algebra routines
                **'dok'** : Enables subscript access of locations
        drop_zeros : bool, default is ``False``
            If ``True``, applies the ``eliminate_zeros`` method of the sparse
            array to remove all zero locations.

        Returns
        -------
        sparse_array
            An incidence matrix in the specified sparse format

        Notes
        -----
        The incidence matrix is a cousin to the adjacency matrix, and used
        by OpenPNM for finding the throats connected to a give pore or set
        of pores. Specifically, an incidence matrix has Np rows and Nt
        columns, and each row represents a pore, containing non-zero
        values at the locations corresponding to the indices of the
        throats connected to that pore. The ``weights`` argument indicates
        what value to place at each location, with the default being 1's
        to simply indicate connections. Another useful option is throat
        indices, such that the data values on each row indicate which
        throats are connected to the pore, though this is redundant as it
        is identical to the locations of non-zeros.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> weights = np.random.rand(pn.num_throats(), ) < 0.5
        >>> im = pn.create_incidence_matrix(weights=weights, fmt='csr')

        """
        # Check if provided data is valid
        if weights is None:
            weights = np.ones((self.Nt,), dtype=int)
        if np.shape(weights)[0] == self.Nt:
            weights = np.append(weights, weights)
        elif np.shape(weights)[0] != 2*self.Nt:
            raise Exception('Received dataset of incorrect length')

        conn = self['throat.conns']
        row = conn[:, 1]
        row = np.append(row, conn[:, 0])
        col = np.arange(self.Nt)
        col = np.append(col, col)

        temp = sprs.coo.coo_matrix((weights, (row, col)), (self.Np, self.Nt))

        if drop_zeros:
            temp.eliminate_zeros()

        # Convert to requested format
        if fmt == 'coo':
            pass  # temp is already in coo format
        elif fmt == 'csr':
            temp = temp.tocsr()
        elif fmt == 'lil':
            temp = temp.tolil()
        elif fmt == 'dok':
            temp = temp.todok()

        return temp

    def find_connected_pores(self, throats=[], flatten=False, mode='or'):
        r"""
        Return a list of pores connected to the given list of throats

        Parameters
        ----------
        throats : array_like
            List of throats numbers
        flatten : bool, optional
            If ``True`` (default) a 1D array of unique pore numbers is
            returned. If ``False`` each location in the the returned
            array contains a sub-arras of neighboring pores for each input
            throat, in the order they were sent.
        mode : str
            Specifies logic to filter the resulting list. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'or'         All neighbors of the input pores. Also accepts 'any'
                         and 'union'.
            'xor'        Only neighbors of one and only one input pore. This
                         is useful for counting the pores that are not shared
                         by any of the input pores. Also accepts
                         'exclusive_or'.
            'xnor'       Neighbors that are shared by two or more input pores.
                         This is equivalent to counting all neighbors
                         with 'or', minus those found with 'xor', and is
                         useful for finding neighbors that the inputs have
                         in common.
            'and'        Only neighbors shared by all input pores. Also
                         accepts 'intersection' and 'all'
            ===========  =====================================================

        Returns
        -------
        1D array (if ``flatten`` is ``True``) or ndarray of arrays (if
        ``flatten`` is ``False``)

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> Ps = pn.find_connected_pores(throats=[0, 1])
        >>> print(Ps)
        [[0 1]
         [1 2]]
        >>> Ps = pn.find_connected_pores(throats=[0, 1], flatten=True)
        >>> print(Ps)
        [0 1 2]

        """
        Ts = self._parse_indices(throats)
        pores = topotools.find_connected_sites(bonds=Ts, network=self,
                                               flatten=flatten, logic=mode)
        return pores

    def find_connecting_throat(self, P1, P2):
        r"""
        Return the throat index connecting pairs of pores.

        Parameters
        ----------
        P1 : array_like
            The indices of the pores whose throats are sought. These can
            be vectors of indices, but must be the same length
        P2 : array_like
            The indices of the pores whose throats are sought. These can
            be vectors of indices, but must be the same length

        Returns
        -------
        list
            Returns a list the same length as P1 (and P2) with the each
            element containing the throat index that connects the
            corresponding pores, or `None`` if pores are not connected.

        Notes
        -----
        The returned list can be converted to an ndarray, which will
        convert the ``None`` values to ``nan``. These can then be found
        using ``numpy.isnan``.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> Ts = pn.find_connecting_throat([0, 1, 2], [2, 2, 2])
        >>> print(Ts)
        [nan  1. nan]

        """
        sites = np.vstack((P1, P2)).T
        Ts = topotools.find_connecting_bonds(sites=sites, network=self)
        return Ts

    def find_neighbor_pores(self, pores, mode='or', flatten=True,
                            include_input=False, asmask=False):
        r"""
        Returns a list of pores that are direct neighbors to the given pore(s)

        Parameters
        ----------
        pores : array_like
            Indices of the pores whose neighbors are sought
        flatten : bool
            If ``True`` (default) the returned result is a compressed
            array of all neighbors.  If ``False``, a list of lists with
            each sub-list containing the neighbors for each input site.
            Note that an *unflattened* list might be slow to generate
            since it is a Python ``list`` rather than a Numpy ``array``.
        include_input : bool
            If ``False`` (default) then the input pores are not included
            in the returned list(s). Note that since pores are not
            neighbors of themselves, the neighbors of pore N will not
            include N, even if this flag is ``True``.
        mode : str
            Specifies logic to filter the resulting list. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'or'         All neighbors of the input pores. Also accepts 'any'
                         and 'union'.
            'xor'        Only neighbors of one and only one input pore. This
                         is useful for counting the pores that are not shared
                         by any of the input pores. Also accepts
                         'exclusive_or'.
            'xnor'       Neighbors that are shared by two or more input pores.
                         This is equivalent to counting all neighbors
                         with 'or', minus those found with 'xor', and is
                         useful for finding neighbors that the inputs have
                         in common.
            'and'        Only neighbors shared by all input pores. Also
                         accepts 'intersection' and 'all'
            ===========  =====================================================

        asmask : boolean
            If ``False`` (default), the returned result is a list of the
            neighboring pores as indices. If ``True``, the returned result is a
            boolean mask. (Useful for labelling)

        Returns
        -------
        If ``flatten`` is ``True``, returns a 1D array of pore indices
        filtered according to the specified mode.  If ``flatten`` is
        ``False``, returns a list of lists, where each list contains the
        neighbors of the corresponding input pores.

        Notes
        -----
        The ``logic`` options are applied to neighboring pores only, thus
        it is not possible to obtain pores that are part of the global set
        but not neighbors. This is because (a) the list of global pores
        might be very large, and (b) it is not possible to return a list
        of neighbors for each input pores if global pores are considered.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> Ps = pn.find_neighbor_pores(pores=[0, 2])
        >>> print(Ps)
        [ 1  3  5  7 25 27]
        >>> Ps = pn.find_neighbor_pores(pores=[0, 1])
        >>> print(Ps)
        [ 2  5  6 25 26]
        >>> Ps = pn.find_neighbor_pores(pores=[0, 1], mode='or',
        ...                             include_input=True)
        >>> print(Ps)
        [ 0  1  2  5  6 25 26]
        >>> Ps = pn.find_neighbor_pores(pores=[0, 2], flatten=False)
        >>> print(Ps[0])
        [ 1  5 25]
        >>> print(Ps[1])
        [ 1  3  7 27]
        >>> Ps = pn.find_neighbor_pores(pores=[0, 2], mode='xnor')
        >>> print(Ps)
        [1]
        >>> Ps = pn.find_neighbor_pores(pores=[0, 2], mode='xor')
        >>> print(Ps)
        [ 3  5  7 25 27]

        """
        pores = self._parse_indices(pores)
        if np.size(pores) == 0:
            return np.array([], ndmin=1, dtype=int)
        neighbors = topotools.find_neighbor_sites(sites=pores, logic=mode,
                                                  network=self,
                                                  flatten=flatten,
                                                  include_input=include_input)
        if asmask is False:
            return neighbors
        elif flatten is True:
            neighbors = self._tomask(element='pore', indices=neighbors)
            return neighbors
        else:
            raise Exception('Cannot create mask on an unflattened output')

    def find_neighbor_throats(self, pores, mode='or', flatten=True,
                              asmask=False):
        r"""
        Returns a list of throats neighboring the given pore(s)

        Parameters
        ----------
        pores : array_like
            Indices of pores whose neighbors are sought
        flatten : bool, optional
            If ``True`` (default) a 1D array of unique throat indices is
            returned. If ``False`` the returned array contains arrays of
            neighboring throat indices for each input pore, in the order
            they were sent.
        mode : str
            Specifies logic to filter the resulting list. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'or'         (default) All neighbors of the input throats. Also
                         accepts 'any' and 'union'.
            'xor'        Only neighbors of one and only one input throats. This
                         is useful for counting the pores that are not shared
                         by any of the input pores. Also accepts
                         'exclusive_or'.
            'xnor'       Neighbors that are shared by two or more input
                         throats. This is equivalent to counting all neighbors
                         with 'or', minus those found with 'xor', and is
                         useful for finding neighbors that the inputs have
                         in common.
            'and'        Only neighbors shared by all input throats. Also
                         accepts 'intersection' and 'all'
            ===========  =====================================================

        asmask : boolean
            If ``False`` (default), the returned result is a list of the
            neighboring throats as indices. If ``True``, the returned result is a
            boolean mask. (Useful for labelling)

        Returns
        -------
        If ``flatten`` is ``True``, returns a 1D array of throat indices
        filtered according to the specified mode.  If ``flatten`` is
        ``False``, returns a list of lists, where each list contains the
        neighbors of the corresponding input pores.

        Notes
        -----
        The ``logic`` options are applied to neighboring bonds only, thus
        it is not possible to obtain bonds that are part of the global set
        but not neighbors. This is because (a) the list of global bonds
        might be very large, and (b) it is not possible to return a list
        of neighbors for each input site if global sites are considered.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> Ts = pn.find_neighbor_throats(pores=[0, 1])
        >>> print(Ts)
        [  0   1 100 101 200 201]
        >>> Ts = pn.find_neighbor_throats(pores=[0, 1], flatten=False)
        >>> print(Ts[0])
        [  0 100 200]
        >>> print(Ts[1])
        [  0   1 101 201]

        """
        pores = self._parse_indices(pores)
        if np.size(pores) == 0:
            return np.array([], ndmin=1, dtype=int)
        if flatten is False:
            neighbors = topotools.find_neighbor_bonds(sites=pores, logic=mode,
                                                      network=self,
                                                      flatten=flatten)
        else:
            neighbors = topotools.find_neighbor_bonds(sites=pores, logic=mode,
                                                      network=self, flatten=True)
        if asmask is False:
            return neighbors
        elif flatten is True:
            neighbors = self._tomask(element='throat', indices=neighbors)
            return neighbors
        else:
            raise Exception('Cannot create mask on an unflattened output')

    def num_neighbors(self, pores, mode='or', flatten=False):
        r"""
        Returns the number of neigbhoring pores for each given input pore

        Parameters
        ----------
        pores : array_like
            Pores whose neighbors are to be counted
        flatten : bool, optional
            If ``False`` (default) the number of pores neighboring each
            input pore as an array the same length as ``pores``. If
            ``True`` the sum total number of is counted.
        mode : str
            The logic to apply to the returned count of pores:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'or'         (default) All neighbors of the input pores. Also
                         accepts 'any' and 'union'.
            'xor'        Only neighbors of one and only one input pore. This
                         is useful for counting the pores that are not shared
                         by any of the input pores. Also accepts
                         'exclusive_or'.
            'xnor'       Neighbors that are shared by two or more input pores.
                         This is equivalent to counting all neighbors
                         with 'or', minus those found with 'xor', and is
                         useful for finding neighbors that the inputs have
                         in common.
            'and'        Only neighbors shared by all input pores. Also
                         accepts 'intersection' and 'all'
            ===========  =====================================================

        Returns
        -------
        If ``flatten`` is False, a 1D array with number of neighbors in each
        element, otherwise a scalar value of the number of neighbors.

        Notes
        -----
        This method literally just counts the number of elements in the array
        returned by ``find_neighbor_pores`` using the same logic.  Explore
        those methods if uncertain about the meaning of the ``mode`` argument
        here.

        See Also
        --------
        find_neighbor_pores
        find_neighbor_throats

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[5, 5, 5])
        >>> Np = pn.num_neighbors(pores=[0, 1], flatten=False)
        >>> print(Np)
        [3 4]
        >>> Np = pn.num_neighbors(pores=[0, 2], flatten=True)
        >>> print(Np)
        6
        >>> Np = pn.num_neighbors(pores=[0, 2], mode='and', flatten=True)
        >>> print(Np)
        1

        """
        pores = self._parse_indices(pores)
        if flatten:
            # Count number of neighbors
            num = self.find_neighbor_pores(pores, flatten=flatten,
                                           mode=mode, include_input=True)
            num = np.size(num)
        else:   # Could be done much faster if flatten == False
            am = self.create_adjacency_matrix(fmt="csr")
            num = am[pores].sum(axis=1).A1
        return num

    def find_nearby_pores(self, pores, r, flatten=False, include_input=False):
        r"""
        Find all pores within a given radial distance of the input pore(s)
        regardless of whether or not they are toplogically connected.

        Parameters
        ----------
        pores : array_like
            The list of pores for whom nearby neighbors are to be found
        r : scalar
            The maximum radius within which the search should be performed
        include_input : bool
            Controls whether the input pores should be included in the
            list of pores nearby the *other pores* in the input list.
            So if ``pores=[1, 2]`` and 1 and 2 are within ``r`` of each
            other, then 1 will be included in the returned for pores
            near 2, and vice-versa *if* this argument is ``True``.
            The default is ``False``.
        flatten : bool
            If ``True`` returns a single list of all pores that match the
            criteria, otherwise returns an array containing a sub-array for
            each input pore, where each sub-array contains the pores that
            are nearby to each given input pore.  The default is False.

        Returns
        -------
            A list of pores which are within the given spatial distance.
            If a list of N pores is supplied, then a an N-long list of
            such lists is returned.  The returned lists each contain the
            pore for which the neighbors were sought.

        Examples
        --------
        >>> import openpnm as op
        >>> pn = op.network.Cubic(shape=[3, 3, 3])
        >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=1)
        >>> print(Ps[0])
        [3 9]
        >>> print(Ps[1])
        [ 2  4 10]
        >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=0.5)
        >>> print(Ps)
        [array([], dtype=int64), array([], dtype=int64)]
        >>> Ps = pn.find_nearby_pores(pores=[0, 1], r=1, flatten=True)
        >>> print(Ps)
        [ 2  3  4  9 10]

        """
        pores = self._parse_indices(pores)
        # Handle an empty array if given
        if np.size(pores) == 0:
            return np.array([], dtype=np.int64)
        if r <= 0:
            raise Exception('Provided distances should be greater than 0')
        # Create kdTree objects
        kd = sptl.cKDTree(self['pore.coords'])
        kd_pores = sptl.cKDTree(self['pore.coords'][pores])
        # Perform search
        Ps_within_r = kd_pores.query_ball_tree(kd, r=r)
        # Remove self from each list
        for i, P in enumerate(Ps_within_r):
            Ps_within_r[i].remove(pores[i])
        # Convert to flattened list by default
        temp = np.concatenate((Ps_within_r))
        Pn = np.unique(temp).astype(np.int64)
        # Remove inputs if necessary
        if include_input is False:
            Pn = Pn[~np.in1d(Pn, pores)]
        # Convert list of lists to a list of ndarrays
        if flatten is False:
            if len(Pn) == 0:  # Deal with no nearby neighbors
                Pn = [np.array([], dtype=np.int64) for i in pores]
            else:
                mask = np.zeros(shape=np.amax((Pn.max(), pores.max())) + 1, dtype=bool)
                mask[Pn] = True
                temp = []
                for item in Ps_within_r:
                    temp.append(np.array(item, dtype=np.int64)[mask[item]])
                Pn = temp
        return Pn

    @property
    def info(self):
        hr = '―' * 78
        errmsg = '--------- ! --------- ! ---------\n'
        pad = 45
        lines = hr + '\n'
        # ---
        lines += 'Number of pores | throats: '.ljust(pad)
        lines += f"{self.Np} | {self.Nt} \n"
        # ---
        lines += 'Pore Coordination (min | mean | max): '.ljust(pad)
        try:
            x = self['pore.coordination_number']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += 'Pore Diameter (min | mean | max): '.ljust(pad)
        try:
            x = self['pore.diameter']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += 'Pore Spacing (min | mean | max): '.ljust(pad)
        try:
            x = self['throat.spacing']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += 'Throat Diameter (min | mean | max): '.ljust(pad)
        try:
            x = self['throat.diameter']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += 'Throat Length (min | mean | max): '.ljust(pad)
        try:
            x = self['throat.length']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += 'Throat Volume (min | mean | max): '.ljust(pad)
        try:
            x = self['throat.volume']
            lines += f"{x.min():.3E} | {x.mean():.3E} | {x.max():.3E} \n"
        except KeyError:
            lines += errmsg
        # ---
        lines += hr
        print(lines)

    @property
    def conns(self):
        r"""Returns the connectivity matrix of the network."""
        return self['throat.conns']

    @property
    def coords(self):
        r"""Returns the list of pore coordinates of the network."""
        return self['pore.coords']



r"""
Collection of ODE integrators for transient simulations
=======================================================

The ``integrators`` module contains wrapper classes for ode solvers.

"""

from ._base import *
from ._scipy import *



import numpy as np
from openpnm.network import Network
from openpnm._skgraph.generators.tools import parse_points
from openpnm._skgraph.generators import voronoi_delaunay_dual
from openpnm.utils import Docorator


__all__ = ['DelaunayVoronoiDual']
docstr = Docorator()


@docstr.dedent
class DelaunayVoronoiDual(Network):
    r"""
    Combined and interconnected Voronoi and Delaunay tessellations

    Parameters
    ----------
    points : array_like or int
        Can either be an N-by-3 array of point coordinates which will be used,
        or a scalar value indicating the number of points to generate
    shape : array_like
        The size and shape of the domain:

        ========== ============================================================
        shape      result
        ========== ============================================================
        [x, y, z]  A 3D cubic domain of dimension x, y and z
        [x, y, 0]  A 2D square domain of size x by y
        ========== ============================================================

    trim : bool, optional
        If ``True`` then all vertices laying outside the domain will
        be removed. This is only useful if ``reflect=True``.
    reflect : bool, optional
        If ``True`` then the base points will be reflected across
        all the faces of the domain prior to performing the tessellation. This
        feature is best combined with ``trim=True``.
    f : float
        The fraction of points which should be reflected.  The default is 1 which
        reflects all the points in the domain, but this can lead to a lot of
        unnecessary points, so setting to 0.1 or 0.2 helps speed, but risks that
        the tessellation may not have smooth faces if not enough points are
        reflected.
    relaxation : int
        The number of time to iteratively relax the base points by moving them to
        the centroid of their respective Voronoi hulls. The default it 0.

    %(Network.parameters)s

    Notes
    -----
    A Delaunay tessellation is performed on a set of base points then the
    corresponding Voronoi diagram is generated.  Finally, each Delaunay node
    is connected to its neighboring Voronoi vertices to create interconnections
    between the two networks.

    All pores and throats are labelled according to their network (i.e.
    'pore.delaunay'), so they can be each assigned to a different Geometry.

    """

    def __init__(
        self,
        shape,
        points,
        trim=True,
        reflect=True,
        f=1,
        relaxation=0,
        **kwargs
    ):
        super().__init__(**kwargs)
        net, vor, tri = voronoi_delaunay_dual(
            shape=shape,
            points=points,
            trim=trim,
            reflect=reflect,
            relaxation=relaxation,
            f=f,
            node_prefix='pore',
            edge_prefix='throat',
        )
        self.update(net)
        self._post_init()
        self.vor = vor
        self.tri = tri

    def find_throat_facets(self, throats=None):
        r"""
        Finds the indicies of the Voronoi nodes that define the facet or
        ridge between the Delaunay nodes connected by the given throat.

        Parameters
        ----------
        throats : array_like
            The throats whose facets are sought.  The given throats should be
            from the 'delaunay' network. If no throats are specified, all
            'delaunay' throats are assumed.

        Notes
        -----
        The method is not well optimized as it scans through each given throat
        inside a for-loop, so it could be slow for large networks.

        """
        if throats is None:
            throats = self.throats('delaunay')
        temp = []
        tvals = self['throat.interconnect'].astype(int)
        am = self.create_adjacency_matrix(weights=tvals, fmt='lil',
                                          drop_zeros=True)
        for t in throats:
            P12 = self['throat.conns'][t]
            Ps = list(set(am.rows[P12][0]).intersection(am.rows[P12][1]))
            temp.append(Ps)
        return np.array(temp, dtype=object)

    def find_pore_hulls(self, pores=None):
        r"""
        Finds the indices of the Voronoi nodes that define the convex hull
        around the given Delaunay nodes.

        Parameters
        ----------
        pores : array_like
            The pores whose convex hull are sought.  The given pores should be
            from the 'delaunay' network.  If no pores are given, then the hull
            is found for all 'delaunay' pores.

        Notes
        -----
        This metod is not fully optimized as it scans through each pore in a
        for-loop, so could be slow for large networks.

        """
        if pores is None:
            pores = self.pores('delaunay')
        temp = []
        tvals = self['throat.interconnect'].astype(int)
        am = self.create_adjacency_matrix(weights=tvals, fmt='lil',
                                          drop_zeros=True)
        for p in pores:
            Ps = am.rows[p]
            temp.append(Ps)
        return np.array(temp, dtype=object)



import logging
from openpnm.io import project_to_dict, _parse_filename
import pandas as pd


logger = logging.getLogger(__name__)


def project_to_hdf5(project, filename=''):
    r"""
    Creates an HDF5 file containing data from the specified objects

    Parameters
    ----------
    network : Network
        The network containing the desired data

    phases : list[Phase]s (optional, default is none)
        A list of phase objects whose data are to be included

    Returns
    -------
    f : hdf5 file handle
        A handle to an hdf5 file.  This must be closed when done (i.e.
        ``f.close()``.
    """
    from h5py import File as hdfFile
    if filename == '':
        filename = project.name
    filename = _parse_filename(filename, ext='hdf')

    dct = project_to_dict(project=project)
    d = pd.json_normalize(dct, sep='.').to_dict(orient='records')[0]
    for k in list(d.keys()):
        d[k.replace('.', '/')] = d.pop(k)
    f = hdfFile(filename, "w")
    for item in list(d.keys()):
        tempname = '_'.join(item.split('.'))
        arr = d[item]
        if d[item].dtype == 'O':
            logger.warning(item + ' has dtype object, will not write to file')
            del d[item]
        elif 'U' in str(arr[0].dtype):
            pass
        else:
            f.create_dataset(name='/'+tempname, shape=arr.shape,
                             dtype=arr.dtype, data=arr)
    return f


def print_hdf5(f, flat=False):
    r"""
    Given an hdf5 file handle, prints to console in a human readable manner

    Parameters
    ----------
    f : file handle
        The hdf5 file to print
    flat : bool
        Flag to indicate if print should be nested or flat.  The default is
        ``flat==False`` resulting in a nested view.
    """
    if flat is False:
        def print_level(f, p='', indent='-'):
            for item in f.keys():
                if hasattr(f[item], 'keys'):
                    p = print_level(f[item], p=p, indent=indent + indent[0])
                elif indent[-1] != ' ':
                    indent = indent + ''
                p = indent + item + '\n' + p
            return p
        p = print_level(f)
        print(p)
    else:
        f.visit(print)



r"""
Phase
-----

This submodule contains models for calculating thermophysical properties of
liquids, gases, and solids.

"""

from ._phasedocs import *
from . import critical_props
from . import density
from . import diffusivity
from . import heat_capacity
from . import misc
from . import mixtures
from . import partition_coefficient
from . import surface_tension
from . import thermal_conductivity
from . import vapor_pressure
from . import viscosity


import logging
logger = logging.getLogger(__name__)


# %%
import inspect as _inspect
import numpy as _np


default_argmap = {
    'T': 'pore.temperature',
    'P': 'pore.pressure',
    'MW': 'param.molecular_weight',
    'Tc': 'param.critical_temperature',
    'Pc': 'param.critical_pressure',
    'Zc': 'param.critical_compressibilty_factor',
    'Vc': 'param.critical_volume',
    'Tm': 'param.melting_temperature',
    'Tb': 'param.boiling_temperature',
    'omega': 'param.acentric_factor',
    'dipole_moment': 'param.dipole_moment',
    'dipole': 'param.dipole_moment',
    'mu': 'pore.viscosity',
    'sigma': 'pore.surface_tension',
    'rhom': 'pore.molar_density',
    'rho': 'pore.density',
}


mixture_argmap = {
    'xs': 'pore.mole_fraction',
    'ys': 'pore.mole_fraction',
    'ks': 'pore.thermal_conductivity.*',
    'ws': 'pore.mass_fraction',
}


def chemicals_wrapper(phase, f, **kwargs):
    r"""
    Wrapper function for calling models in the ``chemicals`` package

    Parameters
    ----------
    phase : dict
        The OpenPNM Species object for which this model should calculate
        values. This object should ideally have all the necessary chemical
        properties in its ``params`` attribute, although this is optional as
        discussed below in relation to ``kwargs``.
    f : function
        The handle of the function to apply, such as
        ``chemicals.viscosity.Letsou_Stiel``.
    kwargs
        By default this function will use
        ``openpnm.models.phase.default_argmap`` to determine which names on
        ``phase`` correspond to each argument required by ``f``. For
        instance, ``default_argmap['Tc'] = 'param.critical_temperature'``
        so any functions that need ``Tc`` will receive
        ``phase['param.critical_temperature']``. Some models only require
        the normal thermodynamic parameters, like ``T``, ``Tc``, and ``Pc``,
        so in many cases no additional arguments are needed beyond whats in the
        default argument map. However, other models require values that are
        themselves calcualted by another model, like ``Cvm``, or perhaps a
        list of constants like ``a0``, ``a1``, etc. It is necessary to provide
        these as keyword arguments, and they will be included in
        ``default_argmap`` as new entries or overwriting existing ones.
        So ``mu='pore.blah'`` will pass ``phase['pore.blah']`` to the
        ``mu`` argument of ``f``. Any arguments ending with an ``s`` are
        assumed to refer to the individual properties of pure components in a
        mixture. So ``mus`` means fetch the viscosity of each component as a
        list, like ``[1.3e-5, 1.9e-5]``. This function will trim the trailing
        ``s`` off any argument name before checking the ``default_argmap`` so
        that the normal pure component values can be looked up.

    Notes
    -----
    This wrapper works with both pure and mixture phases, but the mixture
    models are very slow due to the way ``chemicals`` vectorizes code. For pure
    species it allows the computation of values at many different conditions
    in a vectorized way. The means that the conditions in each pore, such as
    temperature, pressure, etc can be passed and iterpreted as a list of
    conditions. For mixture models, however, the vectorization is done over
    the compositions, at a *fixed* condition. This means that we must do a
    pure-python for-loop (ie. slow) for each individual pore. As such, we
    have re-implemented several of the most useful mixing models offered by
    ``chemicals`` in ``OpenPNM``, and include unit tests to ensure both
    implementations agree.

    """
    import chemicals as _chemicals
    # Update default argmap with any user supplied values
    argmap = default_argmap.copy()
    for k, v in kwargs.items():
        argmap[k] = v
    args = _get_items_from_target(phase, f, argmap)
    # f = getattr(_chemicals.numba_vectorized, f.__name__)
    msg = f"Numba version failed for {f.__name__}, reverting to pure python"
    if len(set(['xs', 'yz', 'zs']).intersection(args.keys())):
        # Call function in for-loop for each pore since they are not vectorized
        logger.info(msg)
        vals = _np.zeros(phase.Np)
        for pore in phase.Ps:
            a = {}
            for item in args.keys():
                if item.endswith('s'):
                    try:
                        a[item] = [args[item][i][pore] for i in range(len(args[item]))]
                    except TypeError:
                        a[item] = args[item]
                else:
                    a[item] = args[item][pore]
            vals[pore] = f(*list(a.values()))
    else:
        try:
            # Get the numba vectorized version of f, or else numpy arrays don't work
            f = getattr(_chemicals.numba_vectorized, f.__name__)
            vals = f(*args.values())
        except AssertionError:
            f = getattr(_chemicals, f.__name__)
            vals = f(*args.values())
            logger.warn(msg)
    return vals


def _get_items_from_target(phase, f, argmap):
    import chemicals as _chemicals
    try:
        arginfo = _inspect.getfullargspec(f)
    except TypeError:
        temp = getattr(_chemicals, f.__name__)
        arginfo = _inspect.getfullargspec(temp)

    # Scan args and pull values from phase
    args = {}
    for item in arginfo.args:
        # Treat mole fraction specially, since it can be xs, ys, or zs
        if item in ['xs', 'ys', 'zs']:
            args[item] = list(phase['pore.mole_fraction'].values())
            continue
        if 'pore.' + item in phase.keys():
            # If eg. pore.Zc was added to dict directly, no argmap needed
            args[item] = phase[f"{'pore.'+item}"]
            continue
        if item in phase.params.keys():
            # If a parameter is in params but not in default_argmap
            args[item] = phase.params[item]
            continue
        if item.endswith('s'):  # Deal specifically with props that end in s
            if item[:-1] in argmap.keys():
                v = list(phase.get_comp_vals(argmap[item[:-1]]).values())
                args[item] = v
            elif item in argmap.keys():
                v = list(phase.get_comp_vals(argmap[item]).values())
                args[item] = v
        else:
            if item in argmap.keys():  # Get basic mixture properties like T&P
                args[item] = phase[argmap[item]]
    return args



import numpy as _np
from openpnm.models import _doctxt


__all__ = ["ad_dif"]


@_doctxt
def ad_dif(
    phase,
    pore_pressure='pore.pressure',
    throat_hydraulic_conductance='throat.hydraulic_conductance',
    throat_diffusive_conductance='throat.diffusive_conductance',
    s_scheme='powerlaw'
):
    r"""
    Calculates the advective-diffusive conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_pressure : str
        %(dict_blurb)s pore pressure
    throat_hydraulic_conductance : str
        %(dict_blurb)s hydraulic conductance
    throat_diffusive_conductance : str
        %(dict_blurb)s throat diffusive conductance
    s_scheme : str
        Name of the space discretization scheme to use

    Returns
    -------
    %(return_arr)s advection-diffuvsion conductance values

    Notes
    -----
    This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the
    end.

    This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area
    can be imposed by passing the proper conduit_shape_factors argument
    when computing the diffusive and hydraulic conductances.

    shape_factor depends on the physics of the problem, i.e.
    diffusion-like processes and fluid flow need different shape factors.

    """
    network = phase.project.network
    cn = network['throat.conns']
    # Find g for half of pore 1, throat, and half of pore 2
    P = phase[pore_pressure]
    gh = phase[throat_hydraulic_conductance]
    gd = phase[throat_diffusive_conductance]
    if gd.size == network.Nt:
        gd = _np.tile(gd, 2)
    # Special treatment when gd is not Nt by 1 (ex. mass partitioning)
    elif gd.size == 2 * network.Nt:
        gd = gd.reshape(network.Nt * 2, order='F')
    else:
        raise Exception(f"Shape of {throat_diffusive_conductance} must either"
                        r" be (Nt,1) or (Nt,2)")

    Qij = -gh * _np.diff(P[cn], axis=1).squeeze()
    Qij = _np.append(Qij, -Qij)

    Peij = Qij / gd
    Peij[(Peij < 1e-10) & (Peij >= 0)] = 1e-10
    Peij[(Peij > -1e-10) & (Peij <= 0)] = -1e-10

    # Correct the flow rate
    Qij = Peij * gd

    if s_scheme == 'upwind':
        w = gd + _np.maximum(0, -Qij)
    elif s_scheme == 'hybrid':
        w = _np.maximum(0, _np.maximum(-Qij, gd - Qij / 2))
    elif s_scheme == 'powerlaw':
        w = gd * _np.maximum(0, (1 - 0.1 * _np.absolute(Peij))**5) + \
            _np.maximum(0, -Qij)
    elif s_scheme == 'exponential':
        w = -Qij / (1 - _np.exp(Peij))
    else:
        raise Exception('Unrecognized discretization scheme: ' + s_scheme)
    w = w.reshape(network.Nt, 2, order='F')
    return w



r"""
Pore Cross-sectional Area Models
--------------------------------

This model contains a selection of functions for Calculating
cross-sectional area assuming the pore body has various shapes.

"""

from ._funcs import *



import logging
import os as os
import numpy as np
from pathlib import Path
from openpnm.network import Network
from openpnm.topotools import trim


logger = logging.getLogger(__name__)


def network_from_marock(filename, voxel_size=1):
    r"""
    Load data from a 3DMA-Rock extracted network.

    Parameters
    ----------
    filename : str
        The location of the 'np2th' and 'th2np' files. This can be an
        absolute path or relative to the current working directory.
    voxel_size : scalar
        The resolution of the image on which 3DMA-Rock was run, in terms of
        the linear length of eac voxel. The default is 1.  This is used to
        scale the voxel counts to actual dimension. It is recommended that
        this value be in SI units [m] to work well with OpenPNM.

    Returns
    -------
    network : dict
        An OpenPNM Network dictionary

    Notes
    -----
    This format consists of two files: 'rockname.np2th' and 'rockname.th2pn'.
    They should be stored together in a folder which is referred to by the
    path argument. These files are binary and therefore not human readable.

    References
    ----------
    3DMA-Rock is a network extraction algorithm developed by Brent Lindquist
    and his group

    It uses Medial Axis thinning to find the skeleton of the pore space, then
    extracts geometrical features such as pore volume and throat
    cross-sectional area.

    [1] Lindquist, W. Brent, S. M. Lee, W. Oh, A. B. Venkatarangan, H. Shin,
        and M. Prodanovic. "3DMA-Rock: A software package for automated analysis
        of rock pore structure in 3-D computed microtomography images." SUNY Stony
        Brook (2005).

    """

    net = {}
    path = Path(filename)
    path = path.resolve()
    for file in os.listdir(path):
        if file.endswith(".np2th"):
            np2th_file = os.path.join(path, file)
        elif file.endswith(".th2np"):
            th2np_file = os.path.join(path, file)

    with open(np2th_file, mode='rb') as f:
        [Np, Nt] = np.fromfile(file=f, count=2, dtype='u4')
        net['pore.boundary_type'] = np.ndarray([Np, ], int)
        net['throat.conns'] = np.ones([Nt, 2], int)*(-1)
        net['pore.coordination'] = np.ndarray([Np, ], int)
        net['pore.ID_number'] = np.ndarray([Np, ], int)
        for i in range(0, Np):
            ID = np.fromfile(file=f, count=1, dtype='u4')
            net['pore.ID_number'][i] = ID
            net['pore.boundary_type'][i] = np.fromfile(file=f, count=1,
                                                       dtype='u1')
            z = np.fromfile(file=f, count=1, dtype='u4')[0]
            net['pore.coordination'][i] = z
            att_pores = np.fromfile(file=f, count=z, dtype='u4')
            att_throats = np.fromfile(file=f, count=z, dtype='u4')
            for j in range(0, len(att_throats)):
                t = att_throats[j] - 1
                p = att_pores[j] - 1
                net['throat.conns'][t] = [i, p]
        net['throat.conns'] = np.sort(net['throat.conns'], axis=1)
        net['pore.volume'] = np.fromfile(file=f, count=Np, dtype='u4')
        nx = np.fromfile(file=f, count=1, dtype='u4')
        nxy = np.fromfile(file=f, count=1, dtype='u4')
        pos = np.fromfile(file=f, count=Np, dtype='u4')
        ny = nxy/nx
        ni = np.mod(pos, nx)
        nj = np.mod(np.floor(pos/nx), ny)
        nk = np.floor(np.floor(pos/nx)/ny)
        net['pore.coords'] = np.array([ni, nj, nk]).T

    with open(th2np_file, mode='rb') as f:
        Nt = np.fromfile(file=f, count=1, dtype='u4')[0]
        net['throat.cross_sectional_area'] = \
            np.ones([Nt, ], dtype=int)*(-1)
        for i in range(0, Nt):
            ID = np.fromfile(file=f, count=1, dtype='u4')
            net['throat.cross_sectional_area'][i] = \
                np.fromfile(file=f, count=1, dtype='f4')
            # numvox = np.fromfile(file=f, count=1, dtype='u4')
            att_pores = np.fromfile(file=f, count=2, dtype='u4')
        nx = np.fromfile(file=f, count=1, dtype='u4')
        nxy = np.fromfile(file=f, count=1, dtype='u4')
        pos = np.fromfile(file=f, count=Nt, dtype='u4')
        ny = nxy/nx
        ni = np.mod(pos, nx)
        nj = np.mod(np.floor(pos/nx), ny)
        nk = np.floor(np.floor(pos/nx)/ny)
        net['throat.coords'] = np.array([ni, nj, nk]).T
        net['pore.internal'] = net['pore.boundary_type'] == 0

    # Convert voxel area and volume to actual dimensions
    net['throat.cross_sectional_area'] = \
        (voxel_size**2)*net['throat.cross_sectional_area']
    net['pore.volume'] = (voxel_size**3)*net['pore.volume']

    network = Network()
    network.update(net)

    # Trim headless throats before returning
    ind = np.where(network['throat.conns'][:, 0] == -1)[0]
    trim(network=network, throats=ind)

    return network



import openpnm.models as mods


cones_and_cylinders = {
    'pore.seed': {
        'model': mods.misc.random,
        'element': 'pore',
        'num_range': [0.2, 0.7],
        'seed': None,
    },
    'pore.max_size': {
        'model': mods.geometry.pore_size.largest_sphere,
        'iters': 10,
    },
    'pore.diameter': {
        'model': mods.misc.product,
        'props': ['pore.max_size', 'pore.seed'],
    },
    'pore.volume': {
        'model': mods.geometry.pore_volume.sphere,
        'pore_diameter': 'pore.diameter',
    },
    'throat.max_size': {
        'model': mods.misc.from_neighbor_pores,
        'mode': 'min',
        'prop': 'pore.diameter',
    },
    'throat.diameter': {
        'model': mods.misc.scaled,
        'factor': 0.5,
        'prop': 'throat.max_size',
    },
    'throat.length': {
        'model': mods.geometry.throat_length.cones_and_cylinders,
        'pore_diameter': 'pore.diameter',
        'throat_diameter': 'throat.diameter',
    },
    'throat.volume': {
        'model': mods.geometry.throat_volume.cylinder,
        'throat_diameter': 'throat.diameter',
        'throat_length': 'throat.length',
    },
    'throat.cross_sectional_area': {
        'model': mods.geometry.throat_cross_sectional_area.cylinder,
        'throat_diameter': 'throat.diameter',
    },
    'throat.diffusive_size_factors': {
        'model': mods.geometry.diffusive_size_factors.cones_and_cylinders,
        'pore_diameter': 'pore.diameter',
        'throat_diameter': 'throat.diameter',
    },
    'throat.hydraulic_size_factors': {
        'model': mods.geometry.hydraulic_size_factors.cones_and_cylinders,
        'pore_diameter': 'pore.diameter',
        'throat_diameter': 'throat.diameter',
    },
}



import numpy as np
from pandas import DataFrame

from openpnm.io import project_to_dict
from openpnm.utils import Project, sanitize_dict


def network_to_pandas(network, join=False, delim='.'):
    """Converts network data to a Pandas DataFrame."""
    proj = Project()
    proj.append(network)
    # Initialize pore and throat data dictionary using Dict class
    pdata = project_to_dict(project=proj, element='pore',
                            flatten=True, categorize_by=[], delim=delim)
    tdata = project_to_dict(project=proj, element='throat',
                            flatten=True, categorize_by=[], delim=delim)
    data = _to_pandas(pdata, tdata, join, Np=network.Np, Nt=network.Nt)
    return data


def project_to_pandas(project, join=False, delim='.'):
    r"""
    Convert the Network and Phase data to a Pandas DataFrame(s)

    Parameters
    ----------
    project : list
        An OpenPNM ``project`` object
    join : bool
        If ``False`` (default), two DataFrames are returned with *pore*
        data in one, and *throat* data in the other.  If ``True`` the pore
        and throat data are combined into a single DataFrame.  This can be
        problematic as it will put NaNs into all the *pore* columns which
        are shorter than the *throat* columns.

    Returns
    -------
    Pandas ``DataFrame`` object(s) containing property and label data in
    each column. If ``join`` was ``False`` (default) the two DataFrames are
    returned in a dict, otherwise a single DataFrame with pore and
    throat data in the same file, despite the column length being
    different.

    """
    network = project.network

    # Initialize pore and throat data dictionary using Dict class
    pdata = project_to_dict(project=project, element='pore',
                            flatten=True, categorize_by=['name'], delim=delim)
    tdata = project_to_dict(project=project, element='throat',
                            flatten=True, categorize_by=['name'], delim=delim)
    data = _to_pandas(pdata, tdata, join, Np=network.Np, Nt=network.Nt)
    return data


def _to_pandas(pdata, tdata, join, Np, Nt):
    # Scan data and convert non-1d arrays to multiple columns
    for key in list(pdata.keys()):
        if np.shape(pdata[key]) != (Np,):
            arr = pdata.pop(key)
            tmp = np.split(arr, arr.shape[1], axis=1)
            cols = range(len(tmp))
            pdata.update({key+'['+str(i)+']': tmp[i].squeeze()
                          for i in cols})
    for key in list(tdata.keys()):
        if np.shape(tdata[key]) != (Nt,):
            arr = tdata.pop(key)
            tmp = np.split(arr, arr.shape[1], axis=1)
            cols = range(len(tmp))
            tdata.update({key+'['+str(i)+']': tmp[i].squeeze()
                          for i in cols})

    # Convert sanitized dictionaries to DataFrames
    pdata = DataFrame(sanitize_dict(pdata))
    tdata = DataFrame(sanitize_dict(tdata))

    # Prepare DataFrames to be returned
    if join:
        data = tdata.join(other=pdata, how='left')
    else:
        data = {'pore': pdata, 'throat': tdata}
    return data



import logging
import numpy as _np
from transforms3d import _gohlketransforms as tr
from openpnm.models import physics as pm
from openpnm.models import _doctxt
from openpnm.models.physics._utils import _get_key_props


logger = logging.getLogger(__name__)


__all__ = [
    "washburn",
    "purcell",
]


@_doctxt
def washburn(phase,
             surface_tension="throat.surface_tension",
             contact_angle="throat.contact_angle",
             diameter="throat.diameter"):
    r"""
    Computes the capillary entry pressure assuming the throat in a
    cylindrical tube.

    Parameters
    ----------
    %(phase)s
    surface_tension : str
        %(dict_blurb)s surface tension. If a pore property is given, it is
        interpolated to a throat list.
    contact_angle : str
        %(dict_blurb)s contact angle. If a pore property is given, it is
        interpolated to a throat list.
    diameter : str
        %(dict_blurb)s throat diameter

    Returns
    -------
    %(return_arr)s capillary entry pressure

    Notes
    -----
    The Washburn equation is:

    .. math::
        P_c = -\frac{2\sigma(cos(\theta))}{r}

    This is the most basic approach to calculating entry pressure and is
    suitable for highly non-wetting invading phases in most materials.

    """
    network = phase.network
    sigma = phase[surface_tension]
    theta = phase[contact_angle]
    r = network[diameter] / 2
    value = -2 * sigma * _np.cos(_np.radians(theta)) / r
    if diameter.split(".")[0] == "throat":
        pass
    else:
        value = value[phase.pores()]
    value[_np.absolute(value) == _np.inf] = 0
    return value


@_doctxt
def purcell(phase,
            r_toroid,
            surface_tension="throat.surface_tension",
            contact_angle="throat.contact_angle",
            diameter="throat.diameter"):
    r"""
    Computes the throat capillary entry pressure assuming the throat is a
    toroid.

    Parameters
    ----------
    %(phase)s
    r_toroid : float or array_like
        The radius of the toroid surrounding the pore
    surface_tension : str
        %(dict_blurb)s surface tension.
    contact_angle : str
        %(dict_blurb)s contact angle.
    diameter : str
        %(dict_blurb)s throat diameter

    Returns
    -------
    %(return_arr)s capillary entry pressure

    Notes
    -----
    This approach accounts for the converging-diverging nature of many throat
    types.  Advancing the meniscus beyond the apex of the toroid requires an
    increase in capillary pressure beyond that for a cylindical tube of the
    same radius. The details of this equation are described by Mason and
    Morrow [1]_, and explored by Gostick [2]_ in the context of a pore network
    model.

    References
    ----------
    .. [1] G. Mason, N. R. Morrow, Effect of contact angle on capillary
           displacement curvatures in pore throats formed by spheres. J.
           Colloid Interface Sci. 168, 130 (1994).
    .. [2] J. Gostick, Random pore network modeling of fibrous PEMFC gas
           diffusion media using Voronoi and Delaunay tessellations. J.
           Electrochem. Soc. 160, F731 (2013).

    """
    network = phase.network
    sigma = phase[surface_tension]
    theta = phase[contact_angle]
    r = network[diameter] / 2
    R = r_toroid
    alpha = (
        theta - 180 + _np.rad2deg(_np.arcsin(_np.sin(_np.radians(theta)) / (1 + r / R)))
    )
    value = (-2 * sigma / r) * (
        _np.cos(_np.radians(theta - alpha))
        / (1 + R / r * (1 - _np.cos(_np.radians(alpha))))
    )
    if diameter.split(".")[0] == "throat":
        pass
    else:
        value = value[phase.pores()]
    return value



import numpy as np
from openpnm import topotools
from openpnm.network import Network
from openpnm.utils import Workspace
from openpnm.utils import Docorator
from openpnm._skgraph.generators import bcc


docstr = Docorator()
ws = Workspace()
__all__ = ['BodyCenteredCubic']


@docstr.dedent
class BodyCenteredCubic(Network):
    r"""
    Body-Centered Cubic lattice which is a simple cubic lattice with additional
    pores located in the center of each cubic unit cell. These pores are
    connected to each other and their diagonal neighbors.

    Parameters
    ----------
    shape : array_like
        The number of pores forming the corners of the unit cell in each
        direction
    spacing : array_like (optional)
        The spacing between pores that form the corners of the unit cells

    %(Network.parameters)s

    See Also
    --------
    Cubic
    FaceCenteredCubic

    Notes
    -----
    The pores are labelled as beloning to 'corner_sites' and 'body_sites' in
    bcc. Throats are labelled by the which type of pores they connect,
    e.g. 'throat.corner_to_body'.

    """

    def __init__(self, shape, mode='sc', spacing=1, **kwargs):
        super().__init__(**kwargs)
        shape = np.array(shape)
        if np.any(shape < 2):
            raise Exception('BCC lattice networks must have at least 2 '
                            'pores in all directions')
        net = bcc(shape=shape, spacing=spacing,
                  node_prefix='pore', edge_prefix='throat')
        self.update(net)
        # Deal with labels
        Ts = self.find_neighbor_throats(pores=self.pores('body'),
                                        mode='exclusive_or')
        self['throat.corner_to_body'] = False
        self['throat.corner_to_body'][Ts] = True
        Ts = self.find_neighbor_throats(pores=self.pores('corner'),
                                        mode='xnor')
        self['throat.corner_to_corner'] = False
        self['throat.corner_to_corner'][Ts] = True
        Ts = self.find_neighbor_throats(pores=self.pores('body'),
                                        mode='xnor')
        self['throat.body_to_body'] = False
        self['throat.body_to_body'][Ts] = True

        # Finally scale network to specified spacing
        topotools.label_faces(self)
        Ps = self.pores(['left', 'right', 'top', 'bottom', 'front', 'back'])
        Ps = self.to_mask(pores=Ps)
        self['pore.surface'] = Ps

    def add_boundary_pores(self, labels, spacing):
        r"""
        Add boundary pores to the specified faces of the network

        Pores are offset from the faces by 1/2 of the given ``spacing``,
        such that they lie directly on the boundaries.

        Parameters
        ----------
        labels : str or list[str]
            The labels indicating the pores defining each face where
            boundary pores are to be added (e.g. 'left' or
            ['left', 'right'])
        spacing : scalar or array_like
            The spacing of the network (e.g. [1, 1, 1]).  This must be
            given since it can be quite difficult to infer from the
            network, for instance if boundary pores have already added
            to other faces.

        """
        spacing = np.array(spacing)
        if spacing.size == 1:
            spacing = np.ones(3)*spacing
        for item in labels:
            Ps = self.pores(item)
            coords = np.absolute(self['pore.coords'][Ps])
            axis = np.count_nonzero(np.diff(coords, axis=0), axis=0) == 0
            offset = np.array(axis, dtype=int)/2
            if np.amin(coords) == np.amin(coords[:, np.where(axis)[0]]):
                offset = -1*offset
            topotools.add_boundary_pores(network=self, pores=Ps, offset=offset,
                                         apply_label=item + '_boundary')



import numpy as _np
from openpnm.models import _doctxt


__all__ = [
    "generic_hydraulic",
    "hagen_poiseuille",
    "valvatne_blunt",
]


@_doctxt
def generic_hydraulic(
    phase,
    pore_viscosity='pore.viscosity',
    throat_viscosity='throat.viscosity',
    size_factors='throat.hydraulic_size_factors'
):
    r"""
    Calculates the hydraulic conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_viscosity : str
        %(dict_blurb)s pore viscosity
    throat_viscosity : str
        %(dict_blurb)s throat viscosity
    size_factors : str
        %(dict_blurb)s conduit hydraulic size factors

    Returns
    -------
    %(return_arr)s hydraulic conductance

    """
    network = phase.network
    conns = network.conns
    mu1, mu2 = phase[pore_viscosity][conns].T
    mut = phase[throat_viscosity]

    SF = network[size_factors]
    if isinstance(SF, dict):  # Legacy approach
        F1, Ft, F2 = SF.values()
    elif SF.ndim > 1:  # Nt-by-3 array
        F1, Ft, F2 = SF.T
    else:  # Nt array, like from network extraction predictions
        F1, Ft, F2 = _np.inf, SF, _np.inf

    g1 = F1 / mu1
    gt = Ft / mut
    g2 = F2 / mu2
    return 1 / (1/g1 + 1/gt + 1/g2)


@_doctxt
def hagen_poiseuille(
    phase,
    pore_viscosity="pore.viscosity",
    throat_viscosity="throat.viscosity",
    size_factors="throat.hydraulic_size_factors"
):
    r"""
    Calculates the hydraulic conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_viscosity : str
        %(dict_blurb)s pore viscosity
    throat_viscosity : str
        %(dict_blurb)s throat viscosity
    size_factors : str
        %(dict_blurb)s conduit hydraulic size factors

    Returns
    -------
    %(return_arr)s hydraulic conductance

    """
    return generic_hydraulic(phase=phase,
                             pore_viscosity=pore_viscosity,
                             throat_viscosity=throat_viscosity,
                             size_factors=size_factors)


@_doctxt
def valvatne_blunt(
    phase,
    pore_viscosity="pore.viscosity",
    throat_viscosity="throat.viscosity",
    pore_shape_factor="pore.shape_factor",
    throat_shape_factor="throat.shape_factor",
    pore_area="pore.area",
    throat_area="throat.cross_sectional_area",
    conduit_lengths="throat.conduit_lengths",
):
    r"""
    Calculates the single phase hydraulic conductance of conduits.

    Function has been adapted for use with the Statoil imported networks
    and makes use of the shape factor in these networks to apply
    Hagen-Poiseuille flow for conduits of different shape classes:
    triangular, square and circular [2].

    Parameters
    ----------
    %(phase)s
    pore_viscosity : str
        %(dict_blurb)s pore viscosity
    throat_viscosity : str
        %(dict_blurb)s throat viscosity
    pore_shape_factor : str
        %(dict_blurb)s pore geometric shape factor
    throat_shape_factor : str
        %(dict_blurb)s throat geometric shape factor
    pore_area : str
        %(dict_blurb)s pore area
        The pore area is calculated using following formula:

        .. math::

            A_P = \frac{R_P^2}{(4 \cdot SF_P)}

        where theoratical value of pore_shape_factor in a circular tube is
        calculated using following formula:

        .. math::

            SF_P = \frac{A_P}{P_P^2} = 1/4π

    throat_area : str
        %(dict_blurb)s throat area.
        The throat area is calculated using following formula:

        .. math::

            T_A = \frac{R_T^2}{(4 \cdot SF_T)}

        where theoratical value of throat shape factor in circular tube is
        calculated using :

        .. math::

            SF_T = \frac{T_A}{T_P^2} = 1/4π

    conduit_lengths : str
        %(dict_blurb)s throat conduit lengths

    Returns
    -------
    %(return_arr)s

    References
    ----------
    [1] Valvatne, Per H., and Martin J. Blunt. "Predictive pore‐scale
    modeling of two‐phase flow in mixed wet media." Water Resources
    Research 40, no. 7 (2004).

    [2] Patzek, T. W., and D. B. Silin (2001), Shape factor and hydraulic
    conductance in noncircular capillaries I. One-phase creeping flow,
    J. Colloid Interface Sci., 236, 295–304.

    """
    network = phase.network
    conns = network["throat.conns"]
    mu_p = phase[pore_viscosity]
    mu_t = phase[throat_viscosity]

    # Fetch model parameters
    L1, Lt, L2 = network[conduit_lengths].T
    Gp = network[pore_shape_factor]
    Gt = network[throat_shape_factor]
    Ap = network[pore_area]
    At = network[throat_area]

    # Throat portions
    tri = Gt <= _np.sqrt(3) / 36.0
    circ = Gt >= 0.07
    square = ~(tri | circ)
    ntri = _np.sum(tri)
    nsquare = _np.sum(square)
    ncirc = _np.sum(circ)
    kt = _np.ones_like(Gt)
    kt[tri] = 3.0 / 5.0
    kt[square] = 0.5623
    kt[circ] = 0.5

    # Pore portions
    tri = Gp <= _np.sqrt(3) / 36.0
    circ = Gp >= 0.07
    square = ~(tri | circ)
    ntri += _np.sum(tri)
    nsquare += _np.sum(square)
    ncirc += _np.sum(circ)
    kp = _np.ones_like(Gp)
    kp[tri] = 3.0 / 5.0
    kp[square] = 0.5623
    kp[circ] = 0.5

    # Calculate conductance values
    gp = kp * Ap**2 * Gp / mu_p
    gt = kt * At**2 * Gt / mu_t
    value = L1 / gp[conns[:, 0]] + Lt / gt + L2 / gp[conns[:, 1]]

    return 1 / value



import logging
import numpy as np
from openpnm.network import Network
from openpnm._skgraph.generators import cubic_template
from openpnm._skgraph.queries import find_coordination
from openpnm._skgraph.tools import dimensionality, find_surface_nodes


__all__ = ['CubicTemplate']


class CubicTemplate(Network):
    r"""
    Simple cubic lattice with arbitrary domain shape specified by a
    template image

    The class creates a standard Cubic network the same shape as the
    provided image, then trims pores from the network that are not in the
    mask.

    Parameters
    ----------
    template : array_like
        The array (image) describing the desired shape of the domain. All
        locations in the image that are marked as ``True`` are kept while
        the rest of trimmed to yeild the shape.
    spacing : array_like, optional
        The spacing between pore centers in each direction. If not given,
        then [1, 1, 1] is assumed.
    name : str
        An optional name for the object to help identify it. If not given,
        one will be generated.

    Notes
    -----
    The other arguments are the same as ``Cubic`` except that ``shape`` is
    inferred from the ``template`` image.

    """

    def __init__(self, template, spacing=[1, 1, 1], label_surface_pores=False, **kwargs):
        super().__init__(**kwargs)
        template = np.atleast_3d(template)
        net = cubic_template(template=template,
                             spacing=spacing,
                             node_prefix='pore',
                             edge_prefix='throat')
        self.update(net)
        if not label_surface_pores:
            return
        self['pore.surface'] = find_surface_nodes(self)
        ndims = dimensionality(self).sum()
        max_neighbors = 6 if ndims == 3 else 4
        num_neighbors = find_coordination(self, nodes=self.Ps)
        mask_internal_surface = (num_neighbors < max_neighbors) & ~self["pore.surface"]
        self.set_label("pore.internal_surface", pores=mask_internal_surface)



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = ["cylinder",
           "cuboid",
           "rectangle"]


@_geodocs
def cylinder(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calcuate the throat perimeter assuming a circular cross-section

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------
    perimeters : ndarray
        A numpy ndarray containing throat perimeter values

    """
    return network[throat_diameter]*_np.pi


@_geodocs
def cuboid(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calcuate the throat perimeter assuming a square cross-section

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------

    """
    return network[throat_diameter]*4


@_geodocs
def rectangle(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calcuate the throat perimeter assuming a rectangular cross-section (2D)

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------

    """
    return 1.0



from openpnm.models.geometry import _geodocs


__all__ = ["spheres_and_cylinders",
           "circles_and_rectangles",
           "cones_and_cylinders",
           "intersecting_cones",
           "hybrid_cones_and_cylinders",
           "trapezoids_and_rectangles",
           "hybrid_trapezoids_and_rectangles",
           "intersecting_trapezoids",
           "pyramids_and_cuboids",
           "intersecting_pyramids",
           "hybrid_pyramids_and_cuboids",
           "cubes_and_cuboids",
           "squares_and_rectangles"]


@_geodocs
def spheres_and_cylinders(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are spheres and throats are
    cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------
    lengths : ndarray
        A numpy ndarray containing throat length values

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.spheres_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def circles_and_rectangles(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter',
):
    r"""
    Finds throat length assuming pores are circles and throats are
    rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.circles_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def cones_and_cylinders(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are cones and throats are
    cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def intersecting_cones(
    network,
    pore_coords="pore.coords",
    throat_coords="throat.coords"
):
    r"""
    Finds throat length assuming pores are intersecting cones.

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.intersecting_cones(
        network=network,
        pore_coords=pore_coords,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def hybrid_cones_and_cylinders(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Finds throat length assuming pores are cones and throats are
    cylinders.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.hybrid_cones_and_cylinders(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def trapezoids_and_rectangles(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are trapezoids and throats are
    rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.trapezoids_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def intersecting_trapezoids(
    network,
    pore_coords="pore.coords",
    throat_coords="throat.coords"
):
    r"""
    Finds throat length assuming pores are intersecting trapezoids.

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.intersecting_trapezoids(
        network=network,
        pore_coords=pore_coords,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def hybrid_trapezoids_and_rectangles(
    network,
    pore_diameter="pore.diameter",
    throat_coords="throat.coords"
):
    r"""
    Finds throat length assuming pores are trapezoids and throats are
    rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Tcoords)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.hybrid_trapezoids_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def pyramids_and_cuboids(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are pyramids and throats are
    cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.pyramids_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def intersecting_pyramids(
    network,
    pore_coords="pore.coords",
    throat_coords="throat.coords"
):
    r"""
    Finds throat length assuming pores are intersecting pyramids.

    Parameters
    ----------
    %(network)s
    %(Pcoords)s
    %(Tcoords)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.intersecting_pyramids(
        network=network,
        pore_coords=pore_coords,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def hybrid_pyramids_and_cuboids(
    network,
    pore_diameter='pore.diameter',
    throat_coords="throat.coords",
):
    r"""
    Finds throat length assuming pores are pyramids and throats are
    cuboids.

    Parameters
    ----------
    %(network)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.hybrid_pyramids_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_coords=throat_coords
    )
    return out[:, 1]


@_geodocs
def cubes_and_cuboids(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are cubes and throats are
    cuboids.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.cubes_and_cuboids(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]


@_geodocs
def squares_and_rectangles(
    network,
    pore_diameter='pore.diameter',
    throat_diameter='throat.diameter'
):
    r"""
    Finds throat length assuming pores are squares and throats are
    rectangles.

    Parameters
    ----------
    %(network)s
    %(Dp)s
    %(Dt)s

    Returns
    -------

    """
    from openpnm.models.geometry import conduit_lengths
    out = conduit_lengths.squares_and_rectangles(
        network=network,
        pore_diameter=pore_diameter,
        throat_diameter=throat_diameter
    )
    return out[:, 1]



from openpnm.models.physics._utils import _poisson_conductance
from openpnm.models import _doctxt


__all__ = ["generic_thermal", "series_resistors"]


@_doctxt
def generic_thermal(phase,
                    pore_conductivity='pore.thermal_conductivity',
                    throat_conductivity='throat.thermal_conductivity',
                    size_factors='throat.diffusive_size_factors'):
    r"""
    Calculate the thermal conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_conductivity : str
        %(dict_blurb)s thermal conductivity
    throat_conductivity : str
        %(dict_blurb)s throat thermal conductivity
    size_factors : str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s thermal conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_conductivity,
                                throat_conductivity=throat_conductivity,
                                size_factors=size_factors)


@_doctxt
def series_resistors(
    phase,
    pore_thermal_conductivity='pore.thermal_conductivity',
    throat_thermal_conductivity='throat.thermal_conductivity',
    size_factors='throat.diffusive_size_factors'
):
    r"""
    Calculate the thermal conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_conductivity : str
        %(dict_blurb)s thermal conductivity
    throat_conductivity : str
        %(dict_blurb)s throat thermal conductivity
    size_factors : str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s thermal conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_thermal_conductivity,
                                throat_conductivity=throat_thermal_conductivity,
                                size_factors=size_factors)



import numpy as np
from openpnm.models.phase.mixtures import mixing_rule
from openpnm.models.phase import _phasedocs


__all__ = [
    'gas_mixture_yweighted',
    'gas_pure_TRC',
    'liquid_pure_rp',
    'liquid_mixture_xweighted',
]


@_phasedocs
def liquid_pure_rp(
    phase,
    T='pore.temperature',
    Tc='param.critical_temperature',
    omega='param.acentric_factor',
    Cpg='pore.heat_capacity_gas',
):
    r"""

    Parameters
    ----------
    %(phase)s
    %(T)s
    %(Tc)s
    %(omega)s
    %(Cpg)s

    """
    # Rowlinson and Poling
    T = phase[T]
    Tc = phase[Tc]
    omega = phase[omega]
    Cpgm = phase[Cpg]
    Tr = T/Tc
    if np.any(Tr > 1):
        raise Exception('Cannot calculate liquid property of fluid above'
                        + 'its critical temperature')
    R = 8.314462618
    lhs = 1.586 + 0.49/(1-Tr) \
        + omega*(4.2775 + 6.3*((1-Tr)**(1/3))/Tr + 0.4355/(1-Tr))
    Cp = lhs*R + Cpgm

    return Cp


@_phasedocs
def gas_pure_TRC(
    phase,
    T='pore.temperature',
    a=[],
):
    r"""

    Parameters
    ----------
    %(phase)s
    %(T)s
    a : list
        The coefficients to use (see notes for form of equation). If not
        given the ``phase['param.CAS']`` is used to lookup the values from
        ``chemicals.heat_capacity.TRC_gas_data``

    Returns
    -------

    """
    # TRCCp
    from chemicals.heat_capacity import TRC_gas_data
    T = phase[T]
    if len(a) == 0:
        c = TRC_gas_data.loc[phase.params['CAS']]
        a = list(c[3:11])
    R = 8.314462618
    y = np.zeros_like(T)
    temp = (T - a[7])/(T + a[6])
    mask = T > a[7]
    y[mask] = temp[mask]
    Cp = R*(a[0] + (a[1]/(T**2))*np.exp(-a[1]/T) + a[3]*(y**2)
            + (a[4] - a[5]/((T - a[7])**2))*(y**8))
    return Cp


@_phasedocs
def gas_mixture_yweighted(
    phase,
    Cps='pore.heat_capacity.*',
):
    r"""
    Uses a linearly mole fraction weighted average

    Parameters
    ----------
    %(phase)s
    %(Cps)s

    Returns
    -------

    """
    Cpmix = mixing_rule(phase=phase, prop=Cps, mode='linear')
    return Cpmix


@_phasedocs
def liquid_mixture_xweighted(
    phase,
    Cps='pore.heat_capacity.*',
):
    r"""
    Uses a linearly mole fraction weighted average

    Parameters
    ----------
    %(phase)s
    %(Cps)s

    Returns
    -------

    """
    Cpmix = mixing_rule(phase=phase, prop=Cps, mode='linear')
    return Cpmix



import logging
import numpy as _np
from openpnm.io import _parse_filename
logger = logging.getLogger(__name__)


def network_to_stl(network, filename=None, maxsize='auto',
                   fileformat='STL Format', logger_level=0):
    r"""
    Saves (transient/steady-state) data from the given objects into the
    specified file.

    Parameters
    ----------
    network : Network.
        The network containing the desired data.
    phases : list[Phase] (place holder, default is None).
        List of phases containing the desired data.
    filename : str (optional).
        The name of the file containing the data to export.
    maxsize : a float or a string "auto" (optional).
        The maximum size of the mesh elements allowed. "auto" corresponds
        to an automatic determination based on pores and throats sizes. Any
        float value will be used as a maximum size. Small values result in
        finner meshes, but slower mesh calculations.
    fileformat : str (optional).
        Default is "STL Format" which corresponds to STL format. Other
        formats such as Gmsh and Fluent are supported (see ngsolve.org).
    logger_level : integer between 0 and 7 (optional).
        Default is 0. The logger level set in netgen package.

    Notes
    -----
    The STL Format is a Standard Triangle (or Tessellation) Language supported
    by many CAD packages and used for 3D printing. Export to this format may be
    slow since a 3D closed surface mesh is built.
    Requires installation of "netgen".
    """
    try:
        import netgen.csg as csg
    except ModuleNotFoundError:
        raise Exception('Module "netgen" not found.')
    try:
        from netgen.meshing import SetMessageImportance as log
        log(logger_level)
    except ModuleNotFoundError:
        logger.warning('Module "netgen.meshing" not found.'
                       + ' The "logger_level" will be ignored.')

    # Temporarily add endpoints to network so STL class works
    network["throat.endpoints.head"] = network.coords[network.conns[:, 0]]
    network["throat.endpoints.tail"] = network.coords[network.conns[:, 1]]

    filename = network.name if filename is None else filename
    path = _parse_filename(filename=filename, ext='stl')
    # Path is a pathlib object, so slice it up as needed
    fname_stl = path.name

    # Correct connections where 'pore.diameter' = 'throat.diameter'
    dt = network['throat.diameter'].copy()
    dp = network['pore.diameter'][network['throat.conns']]
    dt[dp[:, 0] == dt] *= 0.99
    dt[dp[:, 1] == dt] *= 0.99

    scale = max(network['pore.diameter'].max(), dt.max(),
                network['throat.length'].max())
    if maxsize == 'auto':
        maxsize = min(network['pore.diameter'].min(), dt.min(),
                      network['throat.length'].min())
    geo = csg.CSGeometry()

    # Define pores
    geometry = csg.Sphere(csg.Pnt(network['pore.coords'][0, 0]/scale,
                                  network['pore.coords'][0, 1]/scale,
                                  network['pore.coords'][0, 2]/scale),
                          network['pore.diameter'][0]/scale/2)
    for p in range(1, network.Np):
        pore = csg.Sphere(csg.Pnt(network['pore.coords'][p, 0]/scale,
                                  network['pore.coords'][p, 1]/scale,
                                  network['pore.coords'][p, 2]/scale),
                          network['pore.diameter'][p]/scale/2)
        geometry += pore

    # Define throats
    for t in range(network.Nt):
        A = network['throat.endpoints.tail'][t, :]/scale
        B = network['throat.endpoints.head'][t, :]/scale
        V = (B-A)/_np.linalg.norm(B-A)
        plane1 = csg.Plane(csg.Pnt(A[0], A[1], A[2]),
                           csg.Vec(-V[0], -V[1], -V[2]))
        plane2 = csg.Plane(csg.Pnt(B[0], B[1], B[2]),
                           csg.Vec(V[0], V[1], V[2]))
        cylinder = csg.Cylinder(csg.Pnt(A[0], A[1], A[2]),
                                csg.Pnt(B[0], B[1], B[2]),
                                dt[t]/scale/2)
        throat = cylinder * plane1 * plane2
        geometry += throat

    # Add pore and throats to geometry, build mesh, rescale, and export
    geo.Add(geometry)
    mesh = geo.GenerateMesh(maxh=maxsize/scale)
    mesh.Scale(scale)
    mesh.Export(filename=fname_stl, format=fileformat)

    # Remove endpoints label from network
    del network["throat.endpoints"]



from scipy.sparse import csc_matrix, csr_matrix

from openpnm.solvers import DirectSolver

__all__ = ['PardisoSpsolve']


class PardisoSpsolve(DirectSolver):
    """Solves a linear system using ``pypardiso.spsolve``."""

    def solve(self, A, b, **kwargs):
        """Solves the given linear system of equations Ax=b."""
        from pypardiso import spsolve

        if not isinstance(A, (csr_matrix, csc_matrix)):
            A = A.tocsr()
        return (spsolve(A, b), 0)



import numpy as np
from openpnm.models.phase import _phasedocs


__all__ = [
    "mix_and_match",
    "mole_to_mass_fraction",
]


@_phasedocs
def mix_and_match(
    phase,
    prop,
    phases,
    occupancy,
):
    r"""
    Return the given property by looking it up from a list of given phases
    based on occupancy.

    Parameters
    ----------
    %(phase)s
    prop : str
        The dictionary key to the array containing the pore/throat property to
        be used in the calculation.
    phases : list
        List of Phases over which the given `prop` is to be
        averaged out.
    occupancy : str
        The dictionary key to the array containing the occupancy associated
        with each of the given ``phases``.

    Returns
    -------
    weighted_average : ndarray
        Weighted average of the given `prop` averaged over `phases`.

    """
    # Hack for ModelsMixin to not complain (cyclic dep)
    prop = prop.strip("_")
    values = np.zeros_like(phases[0][prop])

    for phase in phases:
        mask = phase[occupancy]
        values[mask] = phase[prop][mask]
    return values


@_phasedocs
def mole_to_mass_fraction(
    phase,
    MWs='param.molecular_weight.*',
):
    r"""
    Convert mole fraction to mass fraction

    Parameters
    ----------
    %(phase)s
    %(MWs)s

    Returns
    -------

    """
    MWs = phase.get_comp_vals(MWs)
    xs = phase['pore.mole_fraction']
    ms = {}
    # Find the actual masses in each pore
    for c in xs.keys():
        ms[c] = xs[c]*MWs[c]
    # Normalize component mass by total mass in each pore
    denom = np.vstack(list(ms.values())).sum(axis=0)
    for c in xs.keys():
        ms[c] = ms[c]/denom
    return ms



import numpy as np
from openpnm.models import _doctxt


__all__ = ["conduit_conductance", "late_filling"]


@_doctxt
def conduit_conductance(phase, throat_conductance,
                        throat_occupancy='throat.occupancy',
                        pore_occupancy='pore.occupancy',
                        mode='strict', factor=1e-6):
    r"""
    Determines the conductance of a pore-throat-pore conduit based on the
    invaded state of each element.

    Parameters
    ----------
    %(phase)s
    throat_conductance : str
        %(dict_blurb)s throat conductance for transport
    pore_occupancy : str
        %(dict_blurb)s pore occupancy of the phase associated with ``phase``.
        An occupancy of 1 means the pore is completely filled with the phase
        and it fully conducts.
    throat_occupancy : str
        %(dict_blurb)s throat occupancy of the phase associated with
        ``phase``. An occupancy of 1 means the pore is completely filled
        with the phase and it fully conducts.
    mode : str
        How agressive the method should be when determining if a conduit is
        closed. Options are:

        ========= ============================================================
        Mode      Description
        ========= ============================================================
        strict    If any pore or throat in the conduit is unoccupied by
                  the given phase, the conduit is closed.
        medium    If either the throat or both pores are unoccupied, the
                  conduit is closed
        loose     Only close the conduit if the throat is unoccupied
        ========= ============================================================

    factor : float (default is 1e-6)
        The factor which becomes multiplied to the original conduit's
        conductance to severely limit transport, but not set it to zero.

    Returns
    -------
    %(return_arr)s adjusted conductance values

    """
    network = phase.network
    Tinv = phase[throat_occupancy] < 0.5
    P12 = network['throat.conns']
    Pinv = phase[pore_occupancy][P12] < 0.5
    if mode == 'loose':
        mask = Tinv
    elif mode == 'medium':
        mask = Tinv + np.all(Pinv, axis=1)
    elif mode == 'strict':
        mask = Tinv + np.any(Pinv, axis=1)
    else:
        raise Exception('Unrecongnized mode '+mode)
    value = phase[throat_conductance].copy()
    value[mask] = value[mask]*factor
    return value


@_doctxt
def late_filling(phase, pressure='pore.pressure',
                 Pc_star='pore.pc_star',
                 Swp_star=0.2, eta=3):
    r"""
    Calculates the fraction of a pore or throat filled with invading fluid
    based on the capillary pressure in the invading phase. The invading phase
    volume is calculated from:

        .. math::
            S_{nwp} = 1 - S_{wp}^{*} (P^{*}/P_{c})^{\eta}

    Parameters
    ----------
    pressure : str
        %(dict_blurb)s capillary pressure in the non-wetting phase (Pc > 0).
    Pc_star : str
        %(dict_blurb)s minimum pressure required to create an interface
        within the pore body or throat.  Typically this would be calculated
        using the Washburn equation.
    Swp_star : float
        The residual wetting phase in an invaded pore or throat at a pressure
        of ``pc_star``.
    eta : float
        Exponent controlling the rate at which wetting phase is displaced with
        increasing pressure.

    Returns
    -------
    %(return_arr)s fraction of each pore or throat that would be filled with
    non-wetting phase at the given phase pressure. This does not account
    for whether or not the element is actually invaded, which requires a
    percolation algorithm of some sort.

    """
    pc_star = phase[Pc_star]
    Pc = phase[pressure]
    # Remove any 0's from the Pc array to prevent numpy div by 0 warning
    Pc = np.maximum(Pc, 1e-9)
    Swp = Swp_star*((pc_star/Pc)**eta)
    values = np.clip(1 - Swp, 0.0, 1.0)
    return values



import logging

import numpy as np

from openpnm.algorithms import Algorithm
from openpnm.algorithms._solution import SolutionContainer, TransientSolution
from openpnm.integrators import ScipyRK45
from openpnm.utils import Docorator

logger = logging.getLogger(__name__)
docstr = Docorator()


__all__ = [
    'TransientMultiPhysics',
]


@docstr.dedent
class TransientMultiPhysicsSettings:
    r"""

    Parameters
    ----------
    %(AlgorithmSettings.parameters)s
    algorithms: list
        List of transient algorithm objects to be solved in a coupled manner

    """
    algorithms = []


@docstr.dedent
class TransientMultiPhysics(Algorithm):
    r"""
    A subclass for transient multiphysics simulations.
    """

    def __init__(self, algorithms, **kwargs):
        super().__init__(**kwargs)
        self.settings._update(TransientMultiPhysicsSettings())
        self.settings.algorithms = [alg.name for alg in algorithms]
        self._algs = algorithms

    def run(self, x0, tspan, saveat=None, integrator=None):
        """
        Runs all of the transient algorithms simultaneoulsy and returns the
        solution.

        Parameters
        ----------
        x0 : ndarray or float
            Array (or scalar) containing initial condition values.
        tspan : array_like
            Tuple (or array) containing the integration time span.
        saveat : array_like or float, optional
            If an array is passed, it signifies the time points at which
            the solution is to be stored, and if a scalar is passed, it
            refers to the interval at which the solution is to be stored.
        integrator : Integrator, optional
            Integrator object which will be used to to the time stepping.
            Can be instantiated using openpnm.integrators module.

        Returns
        -------
        TransientSolution
            The solution object, which is basically a numpy array with
            the added functionality that it can be called to return the
            solution at intermediate times (i.e., those not stored in the
            solution object). In the case of multiphysics, the solution object
            is a combined array of solutions for each physics. The solution
            for each physics is available on each algorithm object
            independently.

        """
        logger.info('Running TransientMultiphysics')
        if np.isscalar(saveat):
            saveat = np.arange(*tspan, saveat)
        if (saveat is not None) and (tspan[1] not in saveat):
            saveat = np.hstack((saveat, [tspan[1]]))
        integrator = ScipyRK45() if integrator is None else integrator
        for i, alg in enumerate(self._algs):
            # Perform pre-solve validations
            alg._validate_settings()
            alg._validate_topology_health()
            alg._validate_linear_system()
            # Write x0 to algorithm the obj (needed by _update_iterative_props)
            x0_i = self._get_x0(x0, i)
            alg['pore.ic'] = x0_i = np.ones(alg.Np, dtype=float) * x0_i
            alg._merge_inital_and_boundary_values()
        # Build RHS (dx/dt = RHS), then integrate the system of ODEs
        rhs = self._build_rhs()
        # Integrate RHS using the given solver
        soln = integrator.solve(rhs, x0, tspan, saveat)
        # Return dictionary containing solution
        self.soln = SolutionContainer()
        for i, alg in enumerate(self._algs):
            # Slice soln and attach as TransientSolution object to each alg
            t = soln.t
            x = soln[i*alg.Np:(i+1)*alg.Np, :]
            alg.soln = TransientSolution(t, x)
            # Add solution of each alg to solution dictionary
            self.soln[alg.settings['quantity']] = alg.soln

    def _run_special(self, x0): ...

    def _build_rhs(self):
        """
        Returns a function handle, which calculates dy/dt = rhs(y, t).

        Notes
        -----
        ``y`` is a composite array that contains ALL the variables that
        the multiphysics algorithm solves for, e.g., if the constituent
        algorithms are ``TransientFickianDiffusion``, and
        ``TransientFourierConduction``, ``y[0:Np-1]`` refers to the
        concentration, and ``[Np:2*Np-1]`` refers to the temperature
        values.

        """
        def ode_func(t, y):
            # Initialize RHS
            rhs = []
            for i, alg in enumerate(self._algs):
                # Get x from y, assume alg.Np is same for all algs
                x = self._get_x0(y, i)  # again use helper function
                # Store x onto algorithm,
                alg.x = x
                # Build A and b
                alg._update_A_and_b()
                A = alg.A.tocsc()
                b = alg.b
                # Retrieve volume
                V = alg.network[alg.settings["pore_volume"]]
                # Calcualte RHS
                rhs_alg = np.hstack(-A.dot(x) + b)/V
                rhs = np.hstack((rhs, rhs_alg))
            return rhs

        return ode_func

    def _get_x0(self, x0, i):
        tmp = [alg.Np for alg in self._algs]
        idx_end = np.cumsum(tmp)
        idx_start = np.hstack((0, idx_end[:-1]))
        x0 = x0[idx_start[i]:idx_end[i]]
        return x0



from .air import air
from .water import water
from .mercury import mercury
from .liquid import standard_liquid_mixture, standard_liquid
from .gas import standard_gas_mixture, standard_gas, binary_gas_mixture



import logging
import numpy as np
from openpnm.network import Network
from openpnm.utils import Docorator
from openpnm import topotools
from openpnm import _skgraph as skgr
skgr.settings.node_prefix = 'pore'
skgr.settings.edge_prefix = 'throat'

docstr = Docorator()
logger = logging.getLogger(__name__)
__all__ = ['Cubic']


@docstr.dedent
class Cubic(Network):
    r"""
    Simple cubic lattice with connectivity from 6 to 26

    Though simple, the Cubic network offers many advantages such as easy
    visualization and accurate determination of domain area and length in
    transport calculations.

    Parameters
    ----------
    shape : array_like
        The [Nx, Ny, Nz] size of the network in terms of the number of pores
        in each direction.  For a 2D network
    spacing : array_like, optional
        The spacing between pore centers in each direction. If not given,
        then [1, 1, 1] is assumed.
    connectivity : int, optional
        The number of connections to neighboring pores.  Connections are made
        symmetrically to any combination of face, edge, or corners neighbors.
        The default is 6 to create a simple cubic structure, but options are:

        ===========  =====================================================
        Value        Result
        ===========  =====================================================
        6            Faces only
        14           Faces and Corners
        18           Faces and Edges
        20           Edges and Corners
        26           Faces, Edges and Corners
        ===========  =====================================================

        For a more random distribution of connectivity, use a high
        ``connectivity`` (i.e. 26) and then delete a fraction of the throats
        using ``openpnm.topotools.reduce_coordination``.  Also note that
        corners-only and edges-only are not permitted since they create
        disconnected networks. If you require one of these topologies you
        can specify 14 or 18, then use ``openpnm.topotools.trim`` to remove
        the face-to-face connections, which can be identified by looking
        for throats with a length equal to the network spacing.

    %(Network.parameters)s

    """

    def __init__(self, shape, spacing=[1, 1, 1], connectivity=6, **kwargs):
        super().__init__(**kwargs)
        net = skgr.generators.cubic(shape=shape, spacing=spacing,
                                    connectivity=connectivity,
                                    node_prefix='pore', edge_prefix='throat')
        self.update(net)
        self._post_init()
        self["pore.surface"] = skgr.tools.find_surface_nodes_cubic(self)
        Ps = self["pore.surface"]
        self["throat.surface"] = np.all(Ps[self["throat.conns"]], axis=1)
        self.update(skgr.generators.tools.label_faces_cubic(self))

    def add_boundary_pores(self, labels=["top", "bottom", "front",
                                         "back", "left", "right"],
                           spacing=None):
        r"""
        Add pores to the faces of the network for use as boundary pores.

        Pores are offset from the faces by 1/2 a lattice spacing such that
        they lie directly on the boundaries.

        Parameters
        ----------
        labels : str or list[str]
            The labels indicating the pores defining each face where
            boundary pores are to be added (e.g. 'left' or
            ['left', 'right'])
        spacing : scalar or array_like
            The spacing of the network (e.g. [1, 1, 1]). This should be
            given since it can be quite difficult to infer from the
            network, for instance if boundary pores have already added to
            other faces.

        """
        if isinstance(labels, str):
            labels = [labels]
        x, y, z = self["pore.coords"].T
        if spacing is None:
            spacing = skgr.tools.get_cubic_spacing(self)
        else:
            spacing = np.array(spacing)
            if spacing.size == 1:
                spacing = np.ones(3) * spacing
        Lcx, Lcy, Lcz = spacing

        offset = {}
        shape = skgr.tools.get_cubic_shape(self)
        offset["front"] = offset["left"] = offset["bottom"] = [0, 0, 0]
        offset["right"] = [Lcx * shape[0], 0, 0]
        offset["back"] = [0, Lcy * shape[1], 0]
        offset["top"] = [0, 0, Lcz * shape[2]]

        scale = {}
        scale["left"] = scale["right"] = [0, 1, 1]
        scale["front"] = scale["back"] = [1, 0, 1]
        scale["bottom"] = scale["top"] = [1, 1, 0]

        for label in labels:
            try:
                Ps = self.pores(label)
                topotools.clone_pores(network=self, pores=Ps,
                                      labels=f"{label}_boundary",
                                      mode='parents')
                # Translate cloned pores
                ind = self.pores(label + "_boundary")
                coords = self["pore.coords"][ind]
                coords = coords * scale[label] + offset[label]
                self["pore.coords"][ind] = coords
            except KeyError:
                msg = f"No pores labelled {label} found, skipping boundary addition"
                logger.warning(msg)



import numpy as _np
import scipy.constants as _const
from openpnm.models import _doctxt
from openpnm.models.physics._utils import _poisson_conductance


__all__ = [
    "generic_diffusive",
    "ordinary_diffusion",
    "mixed_diffusion",
    "taylor_aris_diffusion",
]


@_doctxt
def generic_diffusive(phase,
                      pore_diffusivity="pore.diffusivity",
                      throat_diffusivity="throat.diffusivity",
                      size_factors="throat.diffusive_size_factors"):
    r"""
    Calculates the diffusive conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_diffusivity : str
        %(dict_blurb)s pore diffusivity
    throat_diffusivity : str
        %(dict_blurb)s throat diffusivity
    size_factors : str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s diffusive conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_diffusivity,
                                throat_conductivity=throat_diffusivity,
                                size_factors=size_factors)


@_doctxt
def ordinary_diffusion(phase,
                       pore_diffusivity="pore.diffusivity",
                       throat_diffusivity="throat.diffusivity",
                       size_factors="throat.diffusive_size_factors"):
    r"""
    Calculates the diffusive conductance of conduits in network.

    Parameters
    ----------
    %(phase)s
    pore_diffusivity : str
        %(dict_blurb)s pore diffusivity
    throat_diffusivity : str
        %(dict_blurb)s throat diffusivity
    size_factors: str
        %(dict_blurb)s conduit diffusive size factors

    Returns
    -------
    %(return_arr)s diffusive conductance

    """
    return _poisson_conductance(phase=phase,
                                pore_conductivity=pore_diffusivity,
                                throat_conductivity=throat_diffusivity,
                                size_factors=size_factors)


@_doctxt
def mixed_diffusion(phase,
                    pore_diameter="pore.diameter",
                    throat_diameter="throat.diameter",
                    pore_diffusivity="pore.diffusivity",
                    throat_diffusivity="throat.diffusivity",
                    pore_temperature="pore.temperature",
                    throat_temperature="throat.temperature",
                    molecular_weight="pore.molecular_weight",
                    size_factors="throat.diffusive_size_factors"):
    r"""
    Calculates the diffusive conductance of conduits in network with
    Knudsen correction.

    Parameters
    ----------
    %(phase)s
    pore_diameter : str
        %(dict_blurb)s pore diameter
    throat_diameter : str
        %(dict_blurb)s throat diameter
    pore_diffusivity : str
        %(dict_blurb)s pore diffusivity
    throat_diffusivity : str
        %(dict_blurb)s pore diffusivity
    pore_temperature : str
        %(dict_blurb)s pore temperature
    molecular_weigth : str
        %(dict_blurb)s pore molecular weight
    size_factors : str
        %(dict_blurb)s size factor

    Returns
    -------
    %(return_arr)s diffusive conductance

    """
    # Fetch model parameters
    Dp = phase[pore_diffusivity]
    Dt = phase[throat_diffusivity]
    dp = phase.network[pore_diameter]
    dt = phase.network[throat_diameter]
    MWp = phase[molecular_weight]
    MWt = phase.interpolate_data(propname='throat.'+molecular_weight.split('.', 1)[-1])
    Tp = phase[pore_temperature]
    Tt = phase[throat_temperature]

    # Calculate Knudsen contribution
    DKp = dp/3 * (8*_const.R*Tp / (_const.pi*MWp)) ** 0.5
    DKt = dt/3 * (8*_const.R*Tt / (_const.pi*MWt)) ** 0.5

    # Calculate mixed diffusivity
    Dp_eff = (1/DKp + 1/Dp) ** -1
    Dt_eff = (1/DKt + 1/Dt) ** -1

    return _poisson_conductance(phase=phase,
                                pore_conductivity=Dp_eff,
                                throat_conductivity=Dt_eff,
                                size_factors=size_factors)


@_doctxt
def taylor_aris_diffusion(phase,
                          pore_area="pore.area",
                          throat_area="throat.cross_sectional_area",
                          pore_diffusivity="pore.diffusivity",
                          throat_diffusivity="throat.diffusivity",
                          pore_pressure="pore.pressure",
                          throat_hydraulic_conductance="throat.hydraulic_conductance",
                          size_factors="throat.diffusive_size_factors"):
    r"""
    Calculates the diffusive conductance of conduits in network
    considering the Taylor-Aris effect (effect of flow on diffusion).

    Parameters
    ----------
    %(phase)s
    pore_area : str
        %(dict_blurb)s pore cross-sectional area
    throat_area : str
        %(dict_blurb)s throat cross-sectional area
    pore_diffusivity : str
        %(dict_blurb)s pore diffusivity
    throat_diffusivity : str
        %(dict_blurb)s pore diffusivity
    pore_pressure : str
        %(dict_blurb)s pore pressure
    throat_hydraulic_conductance : str
        %(dict_blurb)s throat hydraulic conductance
    size_factors : str
        %(dict_blurb)s conduit size factors

    Returns
    -------
    %(return_arr)s diffusive conductance

    """
    network = phase.network
    cn = network['throat.conns']
    F = network[size_factors]

    # Fetch model parameters
    A1, A2 = network[pore_area][cn].T
    At = network[throat_area]
    D1, D2 = phase[pore_diffusivity][cn].T
    Dt = phase[throat_diffusivity]
    P = phase[pore_pressure]
    gh = phase[throat_hydraulic_conductance]

    # Calculate Peclet number
    Qt = -gh * _np.diff(P[cn], axis=1).squeeze()
    u1, ut, u2 = [Qt/Ai for Ai in [A1, At, A2]]
    Pe1 = u1 * ((4 * A1 / _np.pi) ** 0.5) / D1
    Pe2 = u2 * ((4 * A2 / _np.pi) ** 0.5) / D2
    Pet = ut * ((4 * At / _np.pi) ** 0.5) / Dt

    if F.ndim > 1:
        g1 = D1 * (1 + Pe1**2 / 192) * F[:, 0]
        gt = Dt * (1 + Pet**2 / 192) * F[:, 1]
        g2 = D2 * (1 + Pe2**2 / 192) * F[:, 2]
        gtot = 1 / (1/g1 + 1/gt + 1/g2)
    else:
        gtot = Dt * (1 + Pet**2 / 192) * F
    return gtot



import numpy as _np
from openpnm.models.geometry import _geodocs


__all__ = ["cylinder",
           "cuboid",
           "extrusion",
           "rectangle"]


@_geodocs
def cylinder(
    network,
    throat_diameter='throat.diameter',
    throat_length='throat.length',
):
    r"""
    Calculate surface area for a cylindrical throat

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Lt)s

    Returns
    -------
    surface_areas : ndarray
        A numpy ndarray containing throat surface area values

    """
    return _np.pi * network[throat_diameter] * network[throat_length]


@_geodocs
def cuboid(
    network,
    throat_diameter='throat.diameter',
    throat_length='throat.length',
):
    r"""
    Calculate surface area for a cuboid throat

    Parameters
    ----------
    %(network)s
    %(Dt)s
    %(Lt)s

    Returns
    -------

    """
    return 4 * network[throat_diameter] * network[throat_length]


@_geodocs
def extrusion(
    network,
    throat_perimeter='throat.perimeter',
    throat_length='throat.length',
):
    r"""
    Calculate surface area for an arbitrary shaped throat give the perimeter
    and length.

    Parameters
    ----------
    %(network)s
    %(Pt)s
    %(Lt)s

    Returns
    -------

    """
    return network[throat_perimeter] * network[throat_length]


@_geodocs
def rectangle(
    network,
    throat_length='throat.length',
):
    r"""
    Calculate surface area for a rectangular throat

    Only suitable for true 2D simulations

    Parameters
    ----------
    %(network)s
    %(Lt)s

    Returns
    -------

    """
    return 2 * network[throat_length]



import functools
import inspect
import os
import warnings
from collections.abc import Iterable

import numpy as np
import numpy.lib.recfunctions as rf
import scipy.sparse as sparse
from docrep import DocstringProcessor

__all__ = [
    'Docorator',
    'PrintableList',
    'PrintableDict',
    'HealthDict',
    'NestedDict',
    'flat_list',
    'sanitize_dict',
    'methods_to_table',
    'models_to_table',
    'ignore_warnings',
    'is_symmetric',
    'is_valid_propname',
    'is_transient',
    'get_mixture_model_args',
    'dict_to_struct',
    'struct_to_dict',
    'get_printable_props',
    'get_printable_labels',
]


class Docorator(DocstringProcessor):
    """OpenPNM's customized docstring processor."""

    __instance__ = None

    def __new__(cls, *args, **kwargs):
        if Docorator.__instance__ is None:
            Docorator.__instance__ = DocstringProcessor()

        # Add custom parameter type sections
        a = DocstringProcessor.param_like_sections
        Docorator.__instance__.param_like_sections = a + []  # ["Attributes", "Settings"]
        # Add custom text type sections
        a = Docorator.__instance__.text_sections
        Docorator.__instance__.text_sections = a + []

        # Create a single list of all section types
        a = Docorator.__instance__.param_like_sections
        b = Docorator.__instance__.text_sections
        Docorator.__instance__.all_sections = a + b

        return Docorator.__instance__


class PrintableList(list):
    r"""
    Simple subclass of ``list`` that has nice printing. Only works flat lists.

    Examples
    --------
    >>> from openpnm.utils import PrintableList
    >>> temp = ['item1', 'item2', 'item3']
    >>> print(PrintableList(temp))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    1     : item1
    2     : item2
    3     : item3
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    Each line contains the result of ``print(item)`` on each item in the list

    """

    def __str__(self):
        horizontal_rule = "―" * 78
        lines = [horizontal_rule]
        self.sort()
        for i, item in enumerate(self):
            lines.append("{0:<5s} : {1}".format(str(i + 1), item))
        lines.append(horizontal_rule)
        return "\n".join(lines)

    # def __repr__(self):  # pragma: no cover
    #     return self.__str__()


class PrintableDict(dict):
    r"""
    Simple subclass of ``dict`` that has nicer printing.

    Examples
    --------
    >>> from openpnm.utils import PrintableDict
    >>> from numpy import array as arr
    >>> d = {'item1': 1, 'item2': '1', 'item3': [1, 1], 'item4': arr([1, 1])}
    >>> print(PrintableDict(d))
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    Key                                 Value
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    item1                               1
    item2                               1
    item3                               [1, 1]
    item4                               (2,)
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    If the item is a Numpy array the value column will contain the items'
    shape, otherwise it will contain the result of ``print(item)``

    """

    def __init__(self, *args, key="Key", value="Value", **kwargs):
        self._value = value
        self._key = key
        super().__init__(*args, **kwargs)

    # def __repr__(self):  # pragma: no cover
    #     return self.__str__()

    def __str__(self):
        header = "―" * 78
        lines = [header, "{0:<35s} {1}".format(self._key, self._value), header]
        for item in list(self.keys()):
            if item.startswith('_'):
                continue
            if isinstance(self[item], np.ndarray):
                lines.append("{0:<35s} {1}".format(item, np.shape(self[item])))
            else:
                lines.append("{0:<35s} {1}".format(item, self[item]))
        lines.append(header)
        return "\n".join(lines)


class NestedDict(dict):
    """Brief explanation of 'NestedDict'"""

    def __init__(self, mapping={}, delimiter="/"):
        super().__init__()
        self.delimiter = delimiter
        self.update(mapping)
        self.unravel()

    def __setitem__(self, key, value):
        path = key.split(self.delimiter, 1)
        if len(path) > 1:
            if path[0] not in self.keys():
                self[path[0]] = NestedDict(delimiter=self.delimiter)
            self[path[0]][path[1]] = value
        else:
            super().__setitem__(key, value)

    def __missing__(self, key):
        self[key] = NestedDict(delimiter=self.delimiter)
        return self[key]

    def unravel(self):
        for item in self.keys():
            self[item] = self.pop(item)

    def to_dict(self, dct=None):
        if dct is None:
            dct = self
        plain_dict = dict()
        for key in dct.keys():
            value = dct[key]
            if hasattr(value, "keys"):
                plain_dict[key] = self.to_dict(value)
            else:
                plain_dict[key] = value
        return plain_dict

    def keys(self, dicts=True, values=True):
        k = list(super().keys())
        new_keys = []
        for item in k:
            if hasattr(self[item], "keys"):
                if dicts:
                    new_keys.append(item)
            else:
                if values:
                    new_keys.append(item)
        return new_keys

    def __str__(self):
        def print_level(self, p="", indent="-"):
            for item in self.keys():
                if hasattr(self[item], "keys"):
                    p = print_level(self[item], p=p, indent=indent + indent[0])
                elif indent[-1] != " ":
                    indent = indent + ""
                p = indent + item + "\n" + p
            return p

        p = print_level(self)
        return p


class HealthDict(PrintableDict):
    r"""
    This class adds a 'health' check to a standard dictionary.

    This check looks into the dict values, and considers empty lists as
    healthy and all else as unhealthy.  If one or more entries is
    'unhealthy' the health method returns False.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _get_health(self):
        health = True
        for item in list(self.keys()):
            try:
                if len(self[item]) > 0:
                    health = False
            except TypeError:
                if self[item]:
                    health = False
        return health

    health = property(fget=_get_health)

    def __bool__(self):
        return self.health


def flat_list(input_list):
    r"""
    Given a list of nested lists of arbitrary depth, returns a single
    level or 'flat' list.
    """
    def _flatten(l):
        for el in l:
            if isinstance(el, Iterable) and not isinstance(el, (str, bytes)):
                yield from _flatten(el)
            else:
                yield el

    return list(_flatten(input_list))


def sanitize_dict(input_dict):
    r"""
    Given a nested dictionary, ensures that all nested dicts are normal
    Python dicts.  This is necessary for pickling, or just converting
    an 'auto-vivifying' dict to something that acts normal.
    """
    plain_dict = dict()
    for key in input_dict.keys():
        value = input_dict[key]
        if hasattr(value, "keys"):
            plain_dict[key] = sanitize_dict(value)
        else:
            plain_dict[key] = value
    return plain_dict


def methods_to_table(obj):
    r"""
    Converts a methods on an object to a ReST compatible table

    Parameters
    ----------
    obj : Base
        Any object that has a methods
    params : bool
        Indicates whether or not to include a list of parameter
        values in the table. Set to False for just a list of models, and
        True for a more verbose table with all parameter values.

    """
    parent = obj.__class__.__mro__[1]
    temp = inspect.getmembers(parent, predicate=inspect.isroutine)
    parent_funcs = [i[0] for i in temp if not i[0].startswith("_")]

    temp = inspect.getmembers(obj.__class__, predicate=inspect.isroutine)
    obj_funcs = [i[0] for i in temp if not i[0].startswith("_")]
    funcs = set(obj_funcs).difference(set(parent_funcs))

    row = "+" + "-" * 22 + "+" + "-" * 49 + "+"
    fmt = "{0:1s} {1:20s} {2:1s} {3:47s} {4:1s}"
    lines = []
    lines.append(row)
    lines.append(fmt.format("|", "Method", "|", "Description", "|"))
    lines.append(row.replace("-", "="))
    for i, item in enumerate(funcs):
        try:
            s = getattr(obj, item).__doc__.strip()
            end = s.find("\n")
            if end > 47:
                s = s[:44] + "..."
            lines.append(fmt.format("|", item, "|", s[:end], "|"))
            lines.append(row)
        except AttributeError:
            pass
    return "\n".join(lines)


def models_to_table(obj, params=True):
    r"""
    Converts a all the models on an object to a ReST compatible table

    Parameters
    ----------
    obj : Base
        Any object that has a ``models`` attribute
    params : bool
        Indicates whether or not to include a list of parameter
        values in the table. Set to False for just a list of models, and
        True for a more verbose table with all parameter values.

    """
    if not hasattr(obj, "models"):
        raise Exception("Received object does not have any models")
    row = "+" + "-" * 4 + "+" + "-" * 22 + "+" + "-" * 18 + "+" + "-" * 26 + "+"
    fmt = "{0:1s} {1:2s} {2:1s} {3:20s} {4:1s} {5:16s} {6:1s} {7:24s} {8:1s}"
    lines = []
    lines.append(row)
    lines.append(
        fmt.format("|", "#", "|", "Property Name", "|", "Parameter", "|", "Value", "|")
    )
    lines.append(row.replace("-", "="))
    for i, item in enumerate(obj.models.keys()):
        prop = item
        if len(prop) > 20:
            prop = item[:17] + "..."
        temp = obj.models[item].copy()
        model = str(temp.pop("model")).split(" ")[1]
        lines.append(
            fmt.format("|", str(i + 1), "|", prop, "|", "model:", "|", model, "|")
        )
        lines.append(row)
        if params:
            for param in temp.keys():
                p1 = param
                if len(p1) > 16:
                    p1 = p1[:13] + "..."
                p2 = str(temp[param])
                if len(p2) > 24:
                    p2 = p2[:21] + "..."
                lines.append(fmt.format("|", "", "|", "", "|", p1, "|", p2, "|"))
                lines.append(row)
    return "\n".join(lines)


def ignore_warnings(warning=RuntimeWarning):
    r"""
    Decorator for catching warnings. Useful in pore-scale models where
    nans are inevitable, and numpy gets annoying by throwing lots of
    RuntimeWarnings.

    Parameters
    ----------
    warning : Warning
        Python warning type that you want to temporarily ignore

    Examples
    --------
    >>> from openpnm.utils import ignore_warnings
    >>> @ignore_warnings()
    ... def myfun(x):
    ...     return 1/x

    >>> import numpy as np
    >>> x = np.arange(5)
    >>> myfun(x)
    array([       inf, 1.        , 0.5       , 0.33333333, 0.25      ])

    """

    def _ignore_warning(function):
        @functools.wraps(function)
        def __ignore_warning(*args, **kwargs):
            with warnings.catch_warnings(record=True):
                # Catch all warnings of this type
                warnings.simplefilter("always", warning)
                # Execute the function
                result = function(*args, **kwargs)
            return result

        return __ignore_warning

    return _ignore_warning


def is_symmetric(a, rtol=1e-10):
    r"""
    Is ``a`` a symmetric matrix?

    Parameters
    ----------
    a : ndarray or sparse matrix
        Object to check for being a symmetric matrix.
    rtol : float
        Relative tolerance with respect to the smallest entry in ``a``
        that is used to determine if ``a`` is symmetric.

    Returns
    -------
    bool
        ``True`` if ``a`` is a symmetric matrix, ``False`` otherwise.

    """
    if not isinstance(a, np.ndarray) and not sparse.issparse(a):
        raise Exception("'a' must be either a sparse matrix or an ndarray.")
    if a.shape[0] != a.shape[1]:
        raise Exception("'a' must be a square matrix.")

    atol = np.amin(np.absolute(a.data)) * rtol
    if sparse.issparse(a):
        issym = False if ((a - a.T) > atol).nnz else True
    elif isinstance(a, np.ndarray):
        issym = False if np.any((a - a.T) > atol) else True

    return issym


def get_mixture_model_args(
    phase,
    composition='xs',
    args={
        'mus': 'pore.viscosity',
        'MWs': 'param.molecular_weight',
    }
):
    r"""
    This is used in tests to run models generically
    """
    from openpnm.models.phase.misc import mole_to_mass_fraction
    vals = {}
    if composition in ['ws']:
        temp = np.vstack(list(mole_to_mass_fraction(phase=phase).values()))[:, 0]
        vals[composition] = temp
    else:
        temp = np.vstack(list(phase['pore.mole_fraction'].values()))[:, 0]
        vals[composition] = temp
    for item in args.keys():
        temp = np.vstack(list(phase.get_comp_vals(args[item]).values()))[:, 0]
        vals[item] = temp
    return vals


def dict_to_struct(d):
    r"""
    Converts a dictionary of numpy arrays to a numpy struct

    Parameters
    ----------
    d : dict
        A dictionary wtih numpy arrays in each key. The arrays must be all
        the same size.

    Returns
    -------
    s : numpy struct
        A numpy struct with the fields or names take from the dictionary keys
    """
    struct = rf.unstructured_to_structured(np.vstack(list(d.values())).T,
                                           names=list(d.keys()))
    return struct


def struct_to_dict(s):
    r"""
    Converts a numpy struct array into a dictionary using the struct labels as
    keys

    Parameters
    ----------
    s : numpy struct
        The struct array

    Returns
    -------
    d : dict
        A dictionary with the struct labels or fields as the keys
    """
    d = {}
    for key in s.dtype.names:
        d[key] = s[key]
    return d


def get_printable_props(item, suffix='', hr=78*'―'):
    r"""
    This function is used by the __str__ methods on all classes to get a
    nicely formatted list of properties on the object.

    Parameters
    ----------
    item : dict
        The OpenPNM dictionary object with each dictionary key containing a
        numpy array
    suffix : str, optional
        If provided, this will be attached to the end of every dictionary
        key so that 'pore.viscosity' becomes 'pore.viscosity.phase_01'.  This
        is a workaround to enhance the printing of component information on
        mixtures.
    hr : str, optional
        The horizontal rule to use between the table heading and body

    Returns
    -------
    table : str
        A formatted string that will output a 78 character wide table when
        printed

    Notes
    -----
    The table returned by this function only contains items that are numerical
    arrays.  Any boolean arrays are ignored.

    See Also
    --------
    get_printable_labels

    """
    if suffix and not suffix.startswith('.'):
        suffix = '.' + suffix
    header = [' ']*78
    header[2] = '#'
    header[5:15] = 'Properties'
    header[-12:] = 'Valid Values'
    lines = ''.join(header) + '\n' + hr
    i = 0
    for k, v in item.items():
        if (v.dtype != bool) and not ('._' in k):
            i += 1
            s = [' ']*78
            s[:3] = str(i+1).rjust(3)
            prop = k + suffix
            s[5:5+len(prop)] = prop
            element = k.split('.', 1)[0]
            nans = np.any(np.isnan(np.atleast_2d(v.T)), axis=0)
            valid = str(np.sum(~nans)) + ' / ' + str(item._count(element))
            s[-20:] = valid.rjust(20)
            a = ''.join(s)
            lines = '\n'.join((lines, a))
    return lines


def get_printable_labels(item, suffix='', hr=78*'―'):
    r"""
    This function is used by the __str__ methods on all classes to get a
    nicely formatted list of labels on the object.

    Parameters
    ----------
    item : dict
        The OpenPNM dictionary object with each dictionary key containing a
        numpy array
    suffix : str, optional
        If provided, this will be attached to the end of every dictionary
        key so that 'pore.viscosity' becomes 'pore.viscosity.phase_01'.  This
        is a workaround to enhance the printing of component information on
        mixtures.
    hr : str, optional
        The horizontal rule to use between the table heading and body

    Returns
    -------
    table : str
        A formatted string that will output a 78 character wide table when
        printed

    Notes
    -----
    The table returned by this function only contains items that boolean
    arrays.  Any numerical arrays are ignored.

    See Also
    --------
    get_printable_props
    """
    if suffix and not suffix.startswith('.'):
        suffix = '.' + suffix
    header = [' ']*78
    header[2] = '#'
    header[5:11] = 'Labels'
    header[-18:] = 'Assigned Locations'
    lines = ''.join(header) + '\n' + hr
    i = 0
    for k, v in item.items():
        if (v.dtype == bool) and not ('._' in k):
            i += 1
            s = [' ']*78
            s[:3] = str(i+1).rjust(3)
            prop = k + suffix
            s[5:5+len(prop)] = prop
            valid = str(np.sum(v))
            s[-12:] = valid.rjust(12)
            a = ''.join(s)
            lines = '\n'.join((lines, a))
    return lines


def is_transient(algorithms):
    # check that algorithms is a list
    if type(algorithms) is not list:
        algorithms = [algorithms]
    # return True if any algorithm is transient
    for alg in algorithms:
        if hasattr(alg, 'soln'):
            soln_type = type(alg.soln[alg.settings['quantity']])
            if 'TransientSolution' in str(soln_type):
                return True
    return False


def is_valid_propname(propname):
    r"""
    Checks if ``propname`` is a valid OpenPNM propname, i.e. starts with
    'pore.' or 'throat.'

    Parameters
    ----------
    propname : str
        Property name to check whether it's a valid OpenPNM propname.

    Returns
    -------
    bool
        Whether or not ``propname`` is a valid name

    """
    if not isinstance(propname, str):
        return False
    temp = propname.split(".")
    if temp[0] not in ["pore", "throat"]:
        return False
    if len(temp) == 1:
        return False
    for field in temp:
        if len(field) == 0:
            return False
    return True


def nbr_to_str(nbr, t_precision=12):
    r"""
    Converts a scalar into a string in scientific (exponential) notation
    without the decimal point.

    Parameters
    ----------
    nbr : scalar
        The number to be converted into a scalar.
    t_precision : integer
        The time precision (number of decimal places). Default value is 12.

    Returns
    -------
    num : str
        The string represenation of the given number in scientific notation
    """
    from decimal import Decimal as dc
    n = int(-dc(str(round(nbr, t_precision))).as_tuple().exponent
            * (round(nbr, t_precision) != int(nbr)))
    nbr_str = (str(int(round(nbr, t_precision)*10**n)) + ('e-'+str(n))*(n != 0))
    return nbr_str



import openpnm.models as mods


mercury = {
    'throat.contact_angle': {
        'model': mods.misc.constant,
        'value': 140,
    },
    'pore.density': {
        'model': mods.misc.linear,
        'prop': 'pore.temperature',
        'b': 14280.9,
        'm': -2.47004,
    },
    'pore.molar_density': {
        'model': mods.phase.density.mass_to_molar,
    },
    'pore.surface_tension': {
        'model': mods.misc.linear,
        'prop': 'pore.temperature',
        'b': 0.56254,
        'm': -0.00028,
    },
    'pore.thermal_conductivity': {
        'model': mods.misc.polynomial,
        'prop': 'pore.temperature',
        'a': [3.98691, 0.0170967, -0.0000063862],
    },
    'pore.viscosity': {
        'model': mods.misc.polynomial,
        'prop': 'pore.temperature',
        'a': [0.00355837, -0.0000100131, 1.23684E-08, -5.1684E-12],
    },
}

# Thermophysical Properties of Materials for Nuclear Engineering:
# IAEA, Vienna, 2008. ISBN 978-92-0-106508-7:



import pyamg
from scipy.sparse import csr_matrix

from ._base import IterativeSolver

__all__ = ['PyamgRugeStubenSolver']


# write a PyamgRugeStubenSolver class
class PyamgRugeStubenSolver(IterativeSolver):
    """Iterative solver based on PyAMG's `ruge_stuben_solver`."""

    def solve(self, A, b, x0=None):
        if not isinstance(A, csr_matrix):
            A = A.tocsr()
        ml = pyamg.ruge_stuben_solver(A)
        return ml.solve(b, x0=x0, tol=self.tol, maxiter=self.maxiter, return_info=True)



from numpy import pi as _pi
from openpnm.models.geometry import _geodocs


__all__ = ["cylinder", "cuboid", "rectangle"]


@_geodocs
def cylinder(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calculate throat cross-sectional area for a cylindrical throat

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------

    """
    diams = network[throat_diameter]
    value = _pi / 4 * diams**2
    return value


@_geodocs
def cuboid(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calculate throat cross-sectional area for a cuboid throat

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------

    """
    diams = network[throat_diameter]
    value = (diams)**2
    return value


@_geodocs
def rectangle(
    network,
    throat_diameter='throat.diameter',
):
    r"""
    Calculate throat cross-sectional area for a rectangular throat

    Parameters
    ----------
    %(network)s
    %(Dt)s

    Returns
    -------

    """
    return network[throat_diameter]



r"""
Conduit Lengths Models
----------------------

This model contains a selection of functions for Calculating conduit lengths
in the network assuming pores and throats have different shapes.

"""
from ._funcs import *



"""
Neighbor Lookups
================

"""
import logging
import numpy as np
logger = logging.getLogger(__name__)


__all__ = ['from_neighbor_throats', 'from_neighbor_pores']


def from_neighbor_throats(target, prop, mode='min', ignore_nans=True):
    r"""
    Adopt a value from the values found in neighboring throats

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : str
        The dictionary key of the array containing the throat property to be
        used in the calculation.
    mode : str
        Controls how the pore property is calculated. The default value is
        'min'. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'min'        Returns the value of the minimum property of the
                         neighboring throats
            'max'        Returns the value of the maximum property of the
                         neighboring throats
            'mean'       Returns the value of the mean property of the
                         neighboring throats
            'sum'        Returns the sum of the property of the neighboring
                         throats
            ===========  =====================================================

    Returns
    -------
    value : ndarray
        Array containing customized values based on those of adjacent throats.

    """
    network = target.network
    data = target[prop]
    nans = np.isnan(data)
    im = network.create_incidence_matrix()
    if mode == 'min':
        if ignore_nans:
            data[nans] = np.inf
        values = np.ones((network.Np, ))*np.inf
        np.minimum.at(values, im.row, data[im.col])
    if mode == 'max':
        if ignore_nans:
            data[nans] = -np.inf
        values = np.ones((network.Np, ))*-np.inf
        np.maximum.at(values, im.row, data[im.col])
    if mode == 'mean':
        if ignore_nans:
            data[nans] = 0
        values = np.zeros((network.Np, ))
        np.add.at(values, im.row, data[im.col])
        counts = np.zeros((network.Np, ))
        np.add.at(counts, im.row, np.ones((network.Nt, ))[im.col])
        if ignore_nans:
            np.subtract.at(counts, im.row, nans[im.col])
        values = values/counts
    if mode == 'sum':
        if ignore_nans:
            data[nans] = 0
        values = np.zeros((network.Np, ))
        np.add.at(values, im.row, data[im.col])
    return values


def from_neighbor_pores(target, prop, mode='min', ignore_nans=True):
    r"""
    Adopt a value based on the values in neighboring pores

    Parameters
    ----------
    target : Base
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.
    prop : str
        The dictionary key to the array containing the pore property to be
        used in the calculation.
    mode : str
        Controls how the pore property is calculated. The default value is
        'min'. Options are:

            ===========  =====================================================
            mode         meaning
            ===========  =====================================================
            'min'        Returns the value of the minimum property of the
                         neighboring pores
            'max'        Returns the value of the maximum property of the
                         neighboring pores
            'mean'       Returns the value of the mean property of the
                         neighboring pores
            'sum'        Returns the sum of the property of the neighrboring
                         pores
            ===========  =====================================================

    ignore_nans : bool (default is ``True``)
        If ``True`` the result will ignore ``nans`` in the neighbors

    Returns
    -------
    value : ndarray
        Array containing customized values based on those of adjacent pores.

    """
    network = target.network
    throats = target.Ts
    P12 = network.find_connected_pores(throats)
    pvalues = target[prop][P12]
    if ignore_nans:
        pvalues = np.ma.MaskedArray(data=pvalues, mask=np.isnan(pvalues))
    try:  # If pvalues is not empty
        if mode == 'min':
            value = np.amin(pvalues, axis=1)
        if mode == 'max':
            value = np.amax(pvalues, axis=1)
        if mode == 'mean':
            value = np.mean(pvalues, axis=1)
        if mode == 'sum':
            value = np.sum(pvalues, axis=1)
    except np.AxisError:  # Handle case of empty pvalues
        value = []
    return np.array(value)



import scipy.spatial as sptl
import scipy.sparse as sprs
import numpy as np
from openpnm._skgraph.generators import cubic
from openpnm._skgraph.tools import tri_to_am


def bcc(shape, spacing=1, mode='kdtree', node_prefix='node', edge_prefix='edge'):
    r"""
    Generate a body-centered cubic lattice

    Parameters
    ----------
    shape : array_like
        The number of corner sites in each direction.  A cubic lattice of
        this size is created and then 'body-centered' nodes are added
        afterward.
    spacing : array_like or float
        The size of a unit cell in each direction. If an scalar is given it is
        applied in all 3 directions.
    mode : str
        Dictate how neighbors are found.  Options are:

        ===============  ======================================================
        mode             meaning
        ===============  ======================================================
        'kdtree'         Uses ``scipy.spatial.KDTree`` to find all neighbors
                         within the unit cell
        'triangulation'  Uses ``scipy.spatial.Delaunay`` to find all neighbors
        ===============  ======================================================

    Returns
    -------
    network : dict
        A dictionary containing 'coords', 'conns' and various boolean labels
        (i.e. 'node.center')

    Notes
    -----
    It is not clear whether KDTree or Delaunay are faster. In fact it is
    surely possible to find the neighbors formulaically but this is not
    implemented yet.

    """
    shape = np.array(shape)
    spacing = np.array(spacing)
    net1 = cubic(shape=shape, spacing=1,
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    net2 = cubic(shape=shape-1, spacing=1,
                 node_prefix=node_prefix, edge_prefix=edge_prefix)
    net2[node_prefix + '.coords'] += 0.5
    crds = np.concatenate(
        (net1[node_prefix + '.coords'],
         net2[node_prefix + '.coords']))
    corner_label = np.concatenate(
        (np.ones(net1[node_prefix + '.coords'].shape[0], dtype=bool),
         np.zeros(net2[node_prefix + '.coords'].shape[0], dtype=bool)))
    body_label = np.concatenate(
        (np.zeros(net1[node_prefix + '.coords'].shape[0], dtype=bool),
         np.ones(net2[node_prefix + '.coords'].shape[0], dtype=bool)))
    if mode.startswith('tri'):
        tri = sptl.Delaunay(points=crds)
        am = tri_to_am(tri)
        conns = np.vstack((am.row, am.col)).T
        # Trim diagonal connections between cubic pores
        L = np.sqrt(np.sum(np.diff(crds[conns], axis=1)**2, axis=2)).flatten()
        conns = conns[L <= 1]
    elif mode.startswith('kd'):
        tree1 = sptl.KDTree(crds)
        # Method 1
        hits = tree1.query_ball_point(crds, r=1)
        # Method 2: Not sure which is faster
        # tree2 = sptl.KDTree(crds)
        # hits = tree1.query_ball_tree(tree1, r=1)
        indices = np.hstack(hits)
        # Convert to CSR matrix
        indptr = [len(i) for i in hits]
        indptr.insert(0, 0)
        indptr = np.cumsum(indptr)
        am = sprs.csr_matrix((np.ones_like(indices), indices, indptr))
        am = sprs.triu(am, k=1)
        am = am.tocoo()
        conns = np.vstack((am.row, am.col)).T

    d = {}
    d[node_prefix+'.coords'] = crds*spacing
    d[edge_prefix+'.conns'] = conns
    d[node_prefix+'.corner'] = corner_label
    d[node_prefix+'.body'] = body_label
    return d


if __name__ == '__main__':
    import openpnm as op
    import matplotlib.pyplot as plt
    from openpnm._skgraph import get_node_prefix, get_edge_prefix
    pn = op.network.Network()
    net = bcc([3, 3, 3], 1, mode='tri')
    net['pore.coords'] = net.pop(get_node_prefix(net) + '.coords')
    net['throat.conns'] = net.pop(get_edge_prefix(net) + '.conns')
    net['pore.corner'] = net.pop(get_node_prefix(net) + '.corner')
    net['pore.body'] = net.pop(get_node_prefix(net) + '.body')
    pn.update(net)
    pn['pore.all'] = np.ones((np.shape(pn.coords)[0]), dtype=bool)
    pn['throat.all'] = np.ones((np.shape(pn.conns)[0]), dtype=bool)
    fig, ax = plt.subplots()
    op.topotools.plot_connections(pn, ax=ax)
    op.topotools.plot_coordinates(pn, ax=ax)



import numpy as np
from numba import njit
from scipy.stats import rankdata


__all__ = [
    'qupc_initialize',
    'qupc_update',
    'qupc_compress',
    'qupc_reduce',
]


@njit
def qupc_initialize(size):
    return np.arange(size, dtype=np.int_)


@njit
def qupc_update(arr, ind, val):
    if ind == val:
        arr[ind] = val
    else:
        # Update array and do path compression simultaneously
        while arr[ind] != val:
            arr[ind] = arr[val]
            ind = val
            val = arr[val]
    return arr


def qupc_compress(arr):
    temp = rankdata(arr, method='dense')
    arr[:] = temp
    arr -= 1
    return arr


@njit
def qupc_reduce(arr):
    for i in range(len(arr)-1, 0, -1):
        arr[i] = arr[arr[i]]
    return arr


if __name__ == '__main__':
    a = qupc_initialize(10)
    qupc_update(a, 4, 2)
    qupc_update(a, 7, 4)
    qupc_update(a, 9, 6)
    qupc_update(a, 6, 2)
    qupc_update(a, 5, 9)
    assert np.all(a == [0, 1, 2, 3, 2, 6, 2, 2, 8, 6])
    qupc_reduce(a)
    assert np.all(a == [0, 1, 2, 3, 2, 2, 2, 2, 8, 2])
    qupc_update(a, 9, 9)
    qupc_update(a, 0, 1)
    qupc_update(a, 8, 0)
    assert np.all(a == [1, 1, 2, 3, 2, 2, 2, 2, 1, 9])
    qupc_reduce(a)
    qupc_compress(a)
    assert np.all(a == [0, 0, 1, 2, 1, 1, 1, 1, 0, 3])
    print(a)



r"""
Throat Capillary Shape Factor Models
------------------------------------

This model contains a selection of functions for calculating capillary shape
factor for throats.

"""

from ._funcs import *



import logging
from copy import deepcopy

from openpnm.utils import PrintableDict

logger = logging.getLogger(__name__)


__all__ = [
    'TypedMixin',
    'TypedSet',
    'TypedList',
    'SettingsAttr',
]


class TypedMixin:
    """Based class for enforcing types on lists and sets."""

    def __init__(self, iterable=[], types=[]):
        self._types = types
        if iterable:
            super().__init__(iterable)
            self._set_types()

    def _get_types(self):
        if not hasattr(self, '_types'):
            self._types = []
        if self._types == []:
            self._types = list(set([type(i) for i in self]))
        return self._types

    def _set_types(self):
        if self._types == []:
            self._types = list(set([type(i) for i in self]))
        else:
            raise Exception("Types have already been defined")

    types = property(fget=_get_types, fset=_set_types)

    def _check_type(self, value):
        if (type(value) not in self.types) and (len(self.types) > 0):
            raise TypeError("This list cannot accept values of type "
                            + f"{type(value)}")


class TypedSet(TypedMixin, set):
    """A set that enforces all elements have the same type."""

    def add(self, item):
        self._check_type(item)
        super().add(item)


class TypedList(TypedMixin, list):
    """A list that enforces all elements have the same type."""

    def __setitem__(self, ind, value):
        self._check_type(value)
        super().__setitem__(ind, value)

    def append(self, value):
        self._check_type(value)
        super().append(value)

    def extend(self, iterable):
        for value in iterable:
            self._check_type(value)
        super().extend(iterable)

    def insert(self, index, value):
        self._check_type(value)
        super().insert(index, value)


class SettingsDict(dict):
    def update(self, d):
        for k, v in d:
            self[k] = v
        if "Parameters" in d.__doc__:
            pass

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value


class SettingsAttr:
    r"""
    A custom data class that holds settings for objects.

    The main function of this custom class is to enforce the datatype of
    values that are assigned to ensure they remain consistent.  For instance
    if ``obj.foo = "bar"``, then ``obj.foo = 456`` will fail.
    """

    def __init__(self, *args):
        for i, item in enumerate(args):
            if i == 0:
                super().__setattr__('__doc__', item.__doc__)
            self._update(item)

    def __setattr__(self, attr, value):
        if hasattr(self, attr):
            # If the the attr is already present, check its type
            if getattr(self, attr) is not None:
                # Ensure the written type is an instance of the existing one
                a = value.__class__.__mro__
                b = getattr(self, attr).__class__.__mro__
                c = object().__class__.__mro__
                check = list(set(a).intersection(set(b)).difference(set(c)))
                if len(check) > 0:  # If they share comment parent class
                    super().__setattr__(attr, value)
                else:  # Otherwise raise an error
                    old = type(getattr(self, attr))
                    new = type(value)
                    raise TypeError(f"Attribute \'{attr}\' can only accept "
                                    + f"values of type {old}, but the recieved"
                                    + f" value was of type {new}")
            else:
                # If the current attr is None, let anything be written
                super().__setattr__(attr, value)
        else:
            # If there is no current attr, let anything be written
            super().__setattr__(attr, value)

    def _update(self, settings, docs=False, override=False):
        if settings is None:
            return
        if isinstance(settings, dict):
            docs = False
            for k, v in settings.items():
                v = deepcopy(v)
                if override:
                    super().__setattr__(k, v)
                else:
                    setattr(self, k, v)
        else:  # Dataclass
            attrs = [i for i in dir(settings) if not i.startswith('_')]
            for k in attrs:
                v = deepcopy(getattr(settings, k))
                if override:
                    super().__setattr__(k, v)
                else:
                    setattr(self, k, v)
        if docs:
            self._getdocs(settings)

    @property
    def _attrs(self):
        a = dir(self)
        b = dir(list())
        attrs = list(set(a).difference(set(b)))
        attrs = [i for i in attrs if not i.startswith('_')]
        attrs = sorted(attrs)
        return attrs

    def _deepcopy(self):
        return deepcopy(self)

    def _getdocs(self, settings):
        super().__setattr__('__doc__', settings.__doc__)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __str__(self):  # pragma: no cover
        d = PrintableDict(key="Settings", value="Values")
        d.update(self.__dict__)
        for item in self.__dir__():
            if not item.startswith('_'):
                d[item] = getattr(self, item)
        return d.__str__()

    def __repr__(self):  # pragma: no cover
        return self.__str__()
